# ActivitySim
# See full license in LICENSE.txt.

import datetime
import glob
import logging
import math
import multiprocessing
import os
import threading
from contextlib import contextmanager

import numpy as np
import pandas as pd

from . import config, mem, tracing, util
from .util import GB

logger = logging.getLogger(__name__)

#
# CHUNK_METHODS and METRICS
#

RSS = "rss"
USS = "uss"
BYTES = "bytes"
HYBRID_RSS = "hybrid_rss"
HYBRID_USS = "hybrid_uss"

METRICS = [RSS, USS, BYTES]
CHUNK_METHODS = [RSS, USS, BYTES, HYBRID_RSS, HYBRID_USS]

USS_CHUNK_METHODS = [USS, HYBRID_USS, BYTES]
DEFAULT_CHUNK_METHOD = HYBRID_USS

"""

The chunk_cache table is a record of the memory usage and observed row_size required for chunking the various models.
The row size differs depending on whether memory usage is calculated by rss, uss, or explicitly allocated bytes.
We record all three during training so the mode can be changed without necessitating retraining.

tag,                               num_rows, rss,   uss,   bytes,    uss_row_size, hybrid_uss_row_size, bytes_row_size
atwork_subtour_frequency.simple,   3498,     86016, 81920, 811536,   24,           232,                 232
atwork_subtour_mode_choice.simple, 704,      20480, 20480, 1796608,  30,           2552,                2552
atwork_subtour_scheduling.tour_1,  701,      24576, 24576, 45294082, 36,           64614,               64614
atwork_subtour_scheduling.tour_n,  3,        20480, 20480, 97734,    6827,         32578,               32578
auto_ownership_simulate.simulate,  5000,     77824, 24576, 1400000,  5,            280,                 280

MODE_RETRAIN
    rebuild chunk_cache table and save/replace in output/cache/chunk_cache.csv
    preforms a complete rebuild of chunk_cache table by doing adaptive chunking starting with based on default initial
    settings (DEFAULT_INITIAL_ROWS_PER_CHUNK) and observing rss, uss, and allocated bytes to compute rows_size.
    This will run somewhat slower than the other modes because of overhead of small first chunk, and possible
    instability in the second chunk due to inaccuracies caused by small initial chunk_size sample

MODE_ADAPTIVE
    Use the existing chunk_cache to determine the sizing for the first chunk for each model, but also use the
    observed row_size to adjust the estimated row_size for subsequent chunks. At the end of hte run, writes the
    updated chunk_cache to the output directory, but doesn't overwrite the 'official' cache file. If the user wishes
    they can replace the chunk_cache with the updated versions but this is not done automatically as it is not clear
    this would be the desired behavior. (Might become clearer over time as this is exercised further.)

MODE_PRODUCTION
    Since overhead changes we don't necessarily want the same number of rows per chunk every time
    but we do use the row_size from cache which we trust is stable
    (the whole point of MODE_PRODUCTION is to avoid the cost of observing overhead)
    which is stored in self.initial_row_size because initial_rows_per_chunk used it for the first chunk

MODE_CHUNKLESS
    Do not do chunking, and also do not check or log memory usage, so ActivitySim can focus on performance
    assuming there is abundant RAM.
"""

MODE_RETRAIN = "training"
MODE_ADAPTIVE = "adaptive"
MODE_PRODUCTION = "production"
MODE_CHUNKLESS = "disabled"
TRAINING_MODES = [MODE_RETRAIN, MODE_ADAPTIVE, MODE_PRODUCTION, MODE_CHUNKLESS]

#
# low level
#
ENABLE_MEMORY_MONITOR = True
MEM_MONITOR_TICK = 1  # in seconds

LOG_SUBCHUNK_HISTORY = False  # only useful for debugging
WRITE_SUBCHUNK_HISTORY = False  # only useful for debugging

DEFAULT_INITIAL_ROWS_PER_CHUNK = (
    100  # fallback for default_initial_rows_per_chunk setting
)


#
# cache and history files
#

CACHE_FILE_NAME = "chunk_cache.csv"
LOG_FILE_NAME = "chunk_history.csv"
OMNIBUS_LOG_FILE_NAME = f"omnibus_{LOG_FILE_NAME}"

C_CHUNK_TAG = "tag"
C_DEPTH = "depth"
C_NUM_ROWS = "num_rows"
C_TIME = "time"

# columns to write to LOG_FILE
CUM_OVERHEAD_COLUMNS = [f"cum_overhead_{m}" for m in METRICS]
CHUNK_HISTORY_COLUMNS = (
    [C_TIME, C_CHUNK_TAG]
    + CUM_OVERHEAD_COLUMNS
    + [C_NUM_ROWS, "row_size", "chunk_size", C_DEPTH, "process", "chunk"]
)

CHUNK_CACHE_COLUMNS = [C_CHUNK_TAG, C_NUM_ROWS] + METRICS

#
# globals
#

SETTINGS = {}
CHUNK_LEDGERS = []
CHUNK_SIZERS = []

ledger_lock = threading.Lock()


def chunk_method():
    method = SETTINGS.get("chunk_method")
    if method is None:
        method = SETTINGS.setdefault(
            "chunk_method", config.setting("chunk_method", DEFAULT_CHUNK_METHOD)
        )
        assert (
            method in CHUNK_METHODS
        ), f"chunk_method setting '{method}' not recognized. Should be one of: {CHUNK_METHODS}"
    return method


def chunk_metric():
    return SETTINGS.setdefault(
        "chunk_metric", USS if chunk_method() in USS_CHUNK_METHODS else "rss"
    )


def chunk_training_mode():
    training_mode = SETTINGS.setdefault(
        "chunk_training_mode", config.setting("chunk_training_mode", MODE_ADAPTIVE)
    )
    if not training_mode:
        training_mode = MODE_CHUNKLESS
    assert (
        training_mode in TRAINING_MODES
    ), f"chunk_training_mode '{training_mode}' not one of: {TRAINING_MODES}"
    return training_mode


def chunk_logging():
    return len(CHUNK_LEDGERS) > 0


def default_initial_rows_per_chunk():
    return SETTINGS.setdefault(
        "default_initial_rows_per_chunk",
        config.setting(
            "default_initial_rows_per_chunk", DEFAULT_INITIAL_ROWS_PER_CHUNK
        ),
    )


def min_available_chunk_ratio():
    return SETTINGS.setdefault(
        "min_available_chunk_ratio", config.setting("min_available_chunk_ratio", 0)
    )


def keep_chunk_logs():
    # if we are overwriting MEM_LOG_FILE then presumably we want to delete any subprocess files
    default = LOG_FILE_NAME == OMNIBUS_LOG_FILE_NAME

    return SETTINGS.setdefault(
        "keep_chunk_logs", config.setting("keep_chunk_logs", default)
    )


def trace_label_for_chunk(trace_label, chunk_size, i):
    # add chunk_num to trace_label
    # if chunk_size > 0:
    #     trace_label = tracing.extend_trace_label(trace_label, f'chunk_{i}')
    return trace_label


def get_base_chunk_size():
    assert len(CHUNK_SIZERS) > 0
    return CHUNK_SIZERS[0].chunk_size


def overhead_for_chunk_method(overhead, method=None):
    """

    return appropriate overhead for row_size calculation based on current chunk_method

    * by ChunkSizer.adaptive_rows_per_chunk to determine observed_row_size based on cum_overhead and cum_rows
    * by ChunkSizer.initial_rows_per_chunk to determine initial row_size using cached_history and current chunk_method
    * by consolidate_logs to add informational row_size column to cache file based on chunk_method for training run

    Parameters
    ----------
    overhead: dict keyed by metric or DataFrame with columns

    Returns
    -------
        chunk_method overhead (possibly hybrid, depending on chunk_method)
    """

    def hybrid(xss, bytes):

        # this avoids pessimistic underchunking on second chunk without pre-existing cache
        # but it tends to overshoot on a trained runs
        # hybrid_overhead =  np.maximum(bytes, (xss + bytes) / 2)

        # this approach avoids problems with pure uss, especially with small chunk sizes (e.g. initial training chunks)
        # as numpy may recycle cached blocks and show no increase in uss even though data was allocated and logged
        hybrid_overhead = np.maximum(bytes, xss)

        return hybrid_overhead

    method = method or chunk_method()

    if method == HYBRID_RSS:
        oh = hybrid(overhead[RSS], overhead[BYTES])
    elif method == HYBRID_USS:
        oh = hybrid(overhead[USS], overhead[BYTES])
    else:
        # otherwise method name is same as metric name
        oh = overhead[method]

    return oh


def consolidate_logs():

    glob_file_name = config.log_file_path(f"*{LOG_FILE_NAME}", prefix=False)
    glob_files = glob.glob(glob_file_name)

    if not glob_files:
        return

    assert chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS), (
        f"shouldn't be any chunk log files when chunk_training_mode"
        f" is {MODE_PRODUCTION} or {MODE_CHUNKLESS}"
    )

    #
    # OMNIBUS_LOG_FILE
    #

    logger.debug(f"chunk.consolidate_logs reading glob {glob_file_name}")
    omnibus_df = pd.concat((pd.read_csv(f, comment="#") for f in glob_files))

    omnibus_df = omnibus_df.sort_values(by=C_TIME)

    # shouldn't have different depths for the same chunk_tag
    multi_depth_chunk_tag = omnibus_df[[C_CHUNK_TAG, C_DEPTH]]
    multi_depth_chunk_tag = multi_depth_chunk_tag[~multi_depth_chunk_tag.duplicated()][
        [C_CHUNK_TAG]
    ]
    multi_depth_chunk_tag = multi_depth_chunk_tag[
        multi_depth_chunk_tag[C_CHUNK_TAG].duplicated()
    ]
    assert (
        len(multi_depth_chunk_tag) == 0
    ), f"consolidate_logs multi_depth_chunk_tags \n{multi_depth_chunk_tag.values}"

    if not keep_chunk_logs():
        util.delete_files(glob_files, "chunk.consolidate_logs")

    log_output_path = config.log_file_path(OMNIBUS_LOG_FILE_NAME, prefix=False)
    logger.debug(f"chunk.consolidate_logs writing omnibus log to {log_output_path}")
    omnibus_df.to_csv(log_output_path, mode="w", index=False)

    #
    # CACHE_FILE
    #

    # write cached chunk_history file for use on subsequent runs, optimized for use by ChunkHistorian
    # Essentially a pandas DataFrame keyed by chunk_tag with columns num_rows, rss, uss, bytes

    omnibus_df = omnibus_df[omnibus_df[C_DEPTH] == 1]
    zero_rows = omnibus_df[C_NUM_ROWS] <= 0
    if zero_rows.any():
        # this should only happen when chunk_log() instantiates the base ChunkSizer.
        # Since chunk_log is not chunked (chunk_size is always 0) there is no need for its history record in the cache
        logger.debug(
            f"consolidate_logs dropping {zero_rows.sum()} rows where {C_NUM_ROWS} == 0"
        )
        omnibus_df = omnibus_df[omnibus_df[C_NUM_ROWS] > 0]

    omnibus_df = omnibus_df[[C_CHUNK_TAG, C_NUM_ROWS] + CUM_OVERHEAD_COLUMNS]

    # aggregate by chunk_tag
    omnibus_df = omnibus_df.groupby(C_CHUNK_TAG).sum().reset_index(drop=False)

    # rename cum_overhead_xxx to xxx
    omnibus_df = omnibus_df.rename(columns={f"cum_overhead_{m}": m for m in METRICS})

    # compute row_size
    num_rows = omnibus_df[C_NUM_ROWS]
    for m in USS_CHUNK_METHODS:
        omnibus_df[f"{m}_row_size"] = np.ceil(
            overhead_for_chunk_method(omnibus_df, m) / num_rows
        ).astype(int)

    omnibus_df = omnibus_df.sort_values(by=C_CHUNK_TAG)

    log_dir_output_path = config.log_file_path(CACHE_FILE_NAME, prefix=False)
    logger.debug(
        f"chunk.consolidate_logs writing omnibus chunk cache to {log_dir_output_path}"
    )
    omnibus_df.to_csv(log_dir_output_path, mode="w", index=False)

    if (chunk_training_mode() == MODE_RETRAIN) or not _HISTORIAN.have_cached_history:

        if config.setting("resume_after"):
            # FIXME
            logger.warning(
                f"Not updating chunk_log cache directory because resume_after"
            )
        else:
            cache_dir_output_path = os.path.join(
                config.get_cache_dir(), CACHE_FILE_NAME
            )
            logger.debug(
                f"chunk.consolidate_logs writing chunk cache to {cache_dir_output_path}"
            )
            omnibus_df.to_csv(cache_dir_output_path, mode="w", index=False)


class ChunkHistorian(object):
    """
    Utility for estimating row_size
    """

    def __init__(self):

        self.chunk_log_path = None
        self.have_cached_history = None
        self.cached_history_df = None

    def load_cached_history(self):

        if chunk_training_mode() == MODE_RETRAIN:
            # don't need cached history if retraining
            return

        if self.have_cached_history is not None:
            # already loaded, nothing to do
            return

        chunk_cache_path = os.path.join(config.get_cache_dir(), CACHE_FILE_NAME)

        logger.debug(
            f"ChunkHistorian load_cached_history chunk_cache_path {chunk_cache_path}"
        )

        if os.path.exists(chunk_cache_path):
            logger.debug(
                f"ChunkHistorian load_cached_history reading cached chunk history from {CACHE_FILE_NAME}"
            )
            df = pd.read_csv(chunk_cache_path, comment="#")

            self.cached_history_df = df

            for c in CHUNK_CACHE_COLUMNS:
                assert (
                    c in df
                ), f"Expected column '{c}' not in chunk_cache: {chunk_cache_path}"

            self.have_cached_history = True
        else:
            self.have_cached_history = False

            if chunk_training_mode() == MODE_CHUNKLESS:
                return

            if chunk_training_mode() == MODE_PRODUCTION:
                # raise RuntimeError(f"chunk_training_mode is {MODE_PRODUCTION} but no chunk_cache: {chunk_cache_path}")

                SETTINGS["chunk_training_mode"] = MODE_RETRAIN
                logger.warning(
                    f"chunk_training_mode is {MODE_PRODUCTION} but no chunk_cache: {chunk_cache_path}"
                )
                logger.warning(
                    f"chunk_training_mode falling back to {chunk_training_mode()}"
                )

    def cached_history_for_chunk_tag(self, chunk_tag):

        history = {}
        self.load_cached_history()

        if self.have_cached_history:

            try:
                df = self.cached_history_df[
                    self.cached_history_df[C_CHUNK_TAG] == chunk_tag
                ]
                if len(df) > 0:

                    if len(df) > 1:
                        # don't expect this, but not fatal
                        logger.warning(
                            f"ChunkHistorian aggregating {len(df)} multiple rows for {chunk_tag}"
                        )

                    # history for this chunk_tag as dict column sums ('num_rows' and cum_overhead for each metric)
                    # {'num_rows: <n>, 'rss': <n>, 'uss': <n>, 'bytes': <n>}
                    history = df.sum().to_dict()

            except Exception as e:
                logger.warning(
                    f"ChunkHistorian Error loading cached history for {chunk_tag}"
                )
                raise e

        return history

    def cached_row_size(self, chunk_tag):

        row_size = 0

        cached_history = self.cached_history_for_chunk_tag(chunk_tag)
        if cached_history:
            cum_overhead = {m: cached_history[m] for m in METRICS}
            num_rows = cached_history[C_NUM_ROWS]

            # initial_row_size based on cum_overhead and rows_processed from chunk_cache
            row_size = math.ceil(overhead_for_chunk_method(cum_overhead) / num_rows)

        return row_size

    def write_history(self, history, chunk_tag):

        assert chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS)

        history_df = pd.DataFrame.from_dict(history)

        # just want the last, most up to date row
        history_df = history_df.tail(1)

        history_df[C_CHUNK_TAG] = chunk_tag
        history_df["process"] = multiprocessing.current_process().name

        history_df = history_df[CHUNK_HISTORY_COLUMNS]

        if self.chunk_log_path is None:
            self.chunk_log_path = config.log_file_path(LOG_FILE_NAME)

        tracing.write_df_csv(
            history_df,
            self.chunk_log_path,
            index_label=None,
            columns=None,
            column_labels=None,
            transpose=False,
        )


_HISTORIAN = ChunkHistorian()


class ChunkLedger(object):
    """ """

    def __init__(self, trace_label, chunk_size, baseline_rss, baseline_uss, headroom):
        self.trace_label = trace_label
        self.chunk_size = chunk_size
        self.headroom = headroom
        self.base_chunk_size = get_base_chunk_size()

        self.tables = {}
        self.hwm_bytes = {"value": 0, "info": f"{trace_label}.init"}
        self.hwm_rss = {"value": baseline_rss, "info": f"{trace_label}.init"}
        self.hwm_uss = {"value": baseline_uss, "info": f"{trace_label}.init"}
        self.total_bytes = 0

    def audit(self, msg, bytes=0, rss=0, uss=0, from_rss_monitor=False):

        assert chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS)

        MAX_OVERDRAFT = 0.2

        if not self.base_chunk_size:
            return

        mem_panic_threshold = self.base_chunk_size * (1 + MAX_OVERDRAFT)
        bytes_panic_threshold = self.headroom + (self.base_chunk_size * MAX_OVERDRAFT)

        if bytes > bytes_panic_threshold:
            logger.warning(
                f"out_of_chunk_memory: "
                f"bytes: {bytes} headroom: {self.headroom} chunk_size: {self.base_chunk_size} {msg}"
            )

        if chunk_metric() == RSS and rss > mem_panic_threshold:
            rss, _ = mem.get_rss(force_garbage_collect=True, uss=False)
            if rss > mem_panic_threshold:
                logger.warning(
                    f"out_of_chunk_memory: "
                    f"rss: {rss} chunk_size: {self.base_chunk_size} {msg}"
                )

        if chunk_metric() == USS and uss > mem_panic_threshold:
            _, uss = mem.get_rss(force_garbage_collect=True, uss=True)
            if uss > mem_panic_threshold:
                logger.warning(
                    f"out_of_chunk_memory: "
                    f"uss: {uss} chunk_size: {self.base_chunk_size} {msg}"
                )

    def close(self):
        logger.debug(f"ChunkLedger.close trace_label: {self.trace_label}")
        logger.debug(
            f"ChunkLedger.close hwm_bytes: {self.hwm_bytes.get('value', 0)} {self.hwm_bytes['info']}"
        )
        logger.debug(
            f"ChunkLedger.close hwm_rss {self.hwm_rss['value']} {self.hwm_rss['info']}"
        )
        logger.debug(
            f"ChunkLedger.close hwm_uss {self.hwm_uss['value']} {self.hwm_uss['info']}"
        )

    def log_df(self, table_name, df):
        def size_it(df):
            if isinstance(df, pd.Series):
                elements = util.iprod(df.shape)
                bytes = 0 if not elements else df.memory_usage(index=True)
            elif isinstance(df, pd.DataFrame):
                elements = util.iprod(df.shape)
                bytes = 0 if not elements else df.memory_usage(index=True).sum()
            elif isinstance(df, np.ndarray):
                elements = util.iprod(df.shape)
                bytes = df.nbytes
            elif isinstance(df, list):
                # dict of series, dataframe, or ndarray (e.g. assign assign_variables target and temp dicts)
                elements = 0
                bytes = 0
                for v in df:
                    e, b = size_it(v)
                    elements += e
                    bytes += b
            elif isinstance(df, dict):
                # dict of series, dataframe, or ndarray (e.g. assign assign_variables target and temp dicts)
                elements = 0
                bytes = 0
                for k, v in df.items():
                    e, b = size_it(v)
                    elements += e
                    bytes += b
            else:
                logger.error(f"size_it unknown type: {type(df)}")
                assert False
            return elements, bytes

        assert chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS)

        if df is None:
            elements, bytes = (0, 0)
            delta_bytes = bytes - self.tables.get(table_name, 0)
            self.tables[table_name] = bytes
        else:
            elements, bytes = size_it(df)
            delta_bytes = bytes - self.tables.get(table_name, 0)
            self.tables[table_name] = bytes

        # shape is informational and only used for logging
        if df is None:
            shape = None
        elif isinstance(df, list):
            shape = f"list({[x.shape for x in df]})"
        elif isinstance(df, dict):
            shape = f"dict({[v.shape for v in df.values()]})"
        else:
            shape = df.shape

        logger.debug(
            f"log_df delta_bytes: {util.INT(delta_bytes).rjust(12)} {table_name} {shape} {self.trace_label}"
        )

        # update current total_bytes count
        self.total_bytes = sum(self.tables.values())

    def check_local_hwm(self, hwm_trace_label, rss, uss, total_bytes):

        assert chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS)

        from_rss_monitor = total_bytes is None

        info = (
            f"rss: {GB(rss)} "
            f"uss: {GB(uss)} "
            f"base_chunk_size: {GB(self.base_chunk_size)} "
            f"op: {hwm_trace_label}"
        )

        if total_bytes:
            info = f"bytes: {GB(total_bytes)} " + info

            if total_bytes > self.hwm_bytes["value"]:
                # total_bytes high water mark
                self.hwm_bytes["value"] = total_bytes
                self.hwm_bytes["info"] = info
                self.audit(hwm_trace_label, bytes=total_bytes)

        if rss > self.hwm_rss["value"]:
            # rss high water mark
            self.hwm_rss["value"] = rss
            self.hwm_rss["info"] = info
            self.audit(hwm_trace_label, rss=rss, from_rss_monitor=from_rss_monitor)

        if uss > self.hwm_uss["value"]:
            # uss high water mark
            self.hwm_uss["value"] = uss
            self.hwm_uss["info"] = info
            self.audit(hwm_trace_label, uss=uss, from_rss_monitor=from_rss_monitor)

        # silently registers global high water mark
        mem.check_global_hwm(RSS, rss, hwm_trace_label)
        mem.check_global_hwm(USS, uss, hwm_trace_label)
        if total_bytes:
            mem.check_global_hwm(BYTES, total_bytes, hwm_trace_label)

    def get_hwm_rss(self):
        with ledger_lock:
            net_rss = self.hwm_rss["value"]
        return net_rss

    def get_hwm_uss(self):
        with ledger_lock:
            net_uss = self.hwm_uss["value"]
        return net_uss

    def get_hwm_bytes(self):
        return self.hwm_bytes["value"]


def log_rss(trace_label, force=False):

    if chunk_training_mode() == MODE_CHUNKLESS:
        # no memory tracing at all in chunkless mode
        return

    assert len(CHUNK_LEDGERS) > 0, f"log_rss called without current chunker."

    hwm_trace_label = f"{trace_label}.log_rss"

    if chunk_training_mode() == MODE_PRODUCTION:
        # FIXME - this trace_memory_info call slows things down a lot so it is turned off for now
        # trace_ticks = 0 if force else mem.MEM_TRACE_TICK_LEN
        # mem.trace_memory_info(hwm_trace_label, trace_ticks=trace_ticks)
        return

    rss, uss = mem.trace_memory_info(hwm_trace_label)

    # check local hwm for all ledgers
    with ledger_lock:
        for c in CHUNK_LEDGERS:
            c.check_local_hwm(hwm_trace_label, rss, uss, total_bytes=None)


def log_df(trace_label, table_name, df):

    if chunk_training_mode() in (MODE_PRODUCTION, MODE_CHUNKLESS):
        return

    assert len(CHUNK_LEDGERS) > 0, f"log_df called without current chunker."

    op = "del" if df is None else "add"
    hwm_trace_label = f"{trace_label}.{op}.{table_name}"

    rss, uss = mem.trace_memory_info(hwm_trace_label)

    cur_chunker = CHUNK_LEDGERS[-1]

    # registers this df and recalc total_bytes
    cur_chunker.log_df(table_name, df)

    total_bytes = sum([c.total_bytes for c in CHUNK_LEDGERS])

    # check local hwm for all ledgers
    with ledger_lock:
        for c in CHUNK_LEDGERS:
            c.check_local_hwm(hwm_trace_label, rss, uss, total_bytes)


class MemMonitor(threading.Thread):
    def __init__(self, trace_label, stop_snooping):
        self.trace_label = trace_label
        self.stop_snooping = stop_snooping
        threading.Thread.__init__(self)

    def run(self):
        log_rss(self.trace_label)
        while not self.stop_snooping.wait(timeout=mem.MEM_SNOOP_TICK_LEN):
            log_rss(self.trace_label)


class ChunkSizer(object):
    """ """

    def __init__(self, chunk_tag, trace_label, num_choosers=0, chunk_size=0):

        self.depth = len(CHUNK_SIZERS) + 1

        if chunk_training_mode() != MODE_CHUNKLESS:
            if chunk_metric() == USS:
                self.rss, self.uss = mem.get_rss(force_garbage_collect=True, uss=True)
            else:
                self.rss, _ = mem.get_rss(force_garbage_collect=True, uss=False)
                self.uss = 0

            if self.depth > 1:
                # nested chunkers should be unchunked
                assert chunk_size == 0

                # if we are in a nested call, then we must be in the scope of active Ledger
                # so any rss accumulated so far should be attributed to the parent active ledger
                assert len(CHUNK_SIZERS) == len(CHUNK_LEDGERS)
                parent = CHUNK_SIZERS[-1]
                assert parent.chunk_ledger is not None

                log_rss(
                    trace_label
                )  # give parent a complementary log_rss reading entering sub context
        else:
            self.rss, self.uss = 0, 0
            chunk_size = 0
            config.override_setting("chunk_size", 0)

        self.chunk_tag = chunk_tag
        self.trace_label = trace_label
        self.chunk_size = chunk_size

        self.num_choosers = num_choosers
        self.rows_processed = 0

        min_chunk_ratio = min_available_chunk_ratio()
        assert (
            0 <= min_chunk_ratio <= 1
        ), f"min_chunk_ratio setting {min_chunk_ratio} is not in range [0..1]"
        self.min_chunk_size = chunk_size * min_chunk_ratio

        self.initial_row_size = 0
        self.rows_per_chunk = 0
        self.chunk_ledger = None
        self.history = {}

        self.cum_rows = 0
        self.cum_overhead = {m: 0 for m in METRICS}

        # if production mode, to reduce volatility, initialize cum_overhead and cum_rows from cache
        if chunk_training_mode() in [MODE_ADAPTIVE, MODE_PRODUCTION]:
            cached_history = _HISTORIAN.cached_history_for_chunk_tag(self.chunk_tag)
            if cached_history:
                self.cum_overhead = {m: cached_history[m] for m in METRICS}
                self.cum_rows = cached_history[C_NUM_ROWS]

                logger.debug(
                    f"{self.trace_label}.ChunkSizer - cached history "
                    f"cum_rows: {self.cum_rows} "
                    f"cum_overhead: {self.cum_overhead} "
                )

        # add self to CHUNK_SIZERS list before setting base_chunk_size (since we might be base chunker)
        CHUNK_SIZERS.append(self)

        self.base_chunk_size = CHUNK_SIZERS[0].chunk_size

        # need base_chunk_size to calc headroom
        self.headroom = self.available_headroom(
            self.uss if chunk_metric() == USS else self.rss
        )

    def close(self):

        if ((self.depth == 1) or WRITE_SUBCHUNK_HISTORY) and (
            chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS)
        ):
            _HISTORIAN.write_history(self.history, self.chunk_tag)

        _chunk_sizer = CHUNK_SIZERS.pop()
        assert _chunk_sizer == self

    def available_headroom(self, xss):

        headroom = self.base_chunk_size - xss

        # adjust deficient headroom to min_chunk_size
        if headroom < self.min_chunk_size:

            if self.base_chunk_size > 0:
                logger.warning(
                    f"Not enough memory for minimum chunk_size without exceeding specified chunk_size. "
                    f"available_headroom: {util.INT(headroom)} "
                    f"min_chunk_size: {util.INT(self.min_chunk_size)} "
                    f"base_chunk_size: {util.INT(self.base_chunk_size)}"
                )

            headroom = self.min_chunk_size

        return headroom

    def initial_rows_per_chunk(self):

        # whatever the TRAINING_MODE, use cache to determine initial_row_size
        # (presumably preferable to default_initial_rows_per_chunk)
        self.initial_row_size = _HISTORIAN.cached_row_size(self.chunk_tag)

        if self.chunk_size == 0:
            rows_per_chunk = self.num_choosers
            estimated_number_of_chunks = 1
            self.initial_row_size = 0
        else:

            # we should be a base chunker
            assert len(CHUNK_LEDGERS) == 0, f"len(CHUNK_LEDGERS): {len(CHUNK_LEDGERS)}"

            if self.initial_row_size > 0:
                max_rows_per_chunk = np.maximum(
                    int(self.headroom / self.initial_row_size), 1
                )
                rows_per_chunk = np.clip(max_rows_per_chunk, 1, self.num_choosers)
                estimated_number_of_chunks = math.ceil(
                    self.num_choosers / rows_per_chunk
                )

                logger.debug(
                    f"{self.trace_label}.initial_rows_per_chunk - initial_row_size: {self.initial_row_size}"
                )
            else:
                # if no initial_row_size from cache, fall back to default_initial_rows_per_chunk
                self.initial_row_size = 0
                rows_per_chunk = min(
                    self.num_choosers, default_initial_rows_per_chunk()
                )
                estimated_number_of_chunks = None

                assert chunk_training_mode() != MODE_PRODUCTION

        # cum_rows is out of phase with cum_overhead
        # since we won't know observed_chunk_size until AFTER yielding the chunk
        self.rows_per_chunk = rows_per_chunk
        self.rows_processed += rows_per_chunk
        self.cum_rows += rows_per_chunk

        logger.debug(
            f"{self.trace_label}.initial_rows_per_chunk - "
            f"rows_per_chunk: {self.rows_per_chunk} "
            f"headroom: {self.headroom} "
            f"initial_row_size: {self.initial_row_size} "
        )

        return rows_per_chunk, estimated_number_of_chunks

    def adaptive_rows_per_chunk(self, i):
        # rows_processed is out of phase with cum_overhead
        # overhead is the actual bytes/rss used top process chooser chunk with prev_rows_per_chunk rows

        prev_rows_per_chunk = self.rows_per_chunk
        prev_rows_processed = self.rows_processed
        prev_cum_rows = self.cum_rows
        prev_headroom = self.headroom

        prev_rss = self.rss
        prev_uss = self.uss

        if chunk_training_mode() != MODE_PRODUCTION:

            if chunk_metric() == USS:
                self.rss, self.uss = mem.get_rss(force_garbage_collect=True, uss=True)
            else:
                self.rss, _ = mem.get_rss(force_garbage_collect=True, uss=False)
                self.uss = 0

        self.headroom = self.available_headroom(
            self.uss if chunk_metric() == USS else self.rss
        )

        rows_remaining = self.num_choosers - prev_rows_processed

        if chunk_training_mode() == MODE_PRODUCTION:
            # since overhead changes we don't necessarily want the same number of rows per chunk every time
            # but we do use the row_size from cache which we trust is stable
            # which is stored in self.initial_row_size because initial_rows_per_chunk used it for the first chunk
            observed_row_size = self.initial_row_size
            overhead = self.cum_overhead.copy()
        else:

            # calculate overhead for this chunk iteration
            overhead = {}
            overhead[BYTES] = self.chunk_ledger.get_hwm_bytes()
            overhead[RSS] = self.chunk_ledger.get_hwm_rss() - prev_rss
            overhead[USS] = self.chunk_ledger.get_hwm_uss() - prev_uss

            for m in METRICS:
                self.cum_overhead[m] += overhead[m]

            observed_row_size = prev_cum_rows and math.ceil(
                overhead_for_chunk_method(self.cum_overhead) / prev_cum_rows
            )

        # rows_per_chunk is closest number of chooser rows to achieve chunk_size without exceeding it
        if observed_row_size > 0:
            self.rows_per_chunk = int(self.headroom / observed_row_size)
        else:
            # they don't appear to have used any memory; increase cautiously in case small sample size was to blame
            self.rows_per_chunk = 2 * prev_rows_per_chunk

        self.rows_per_chunk = np.clip(self.rows_per_chunk, 1, rows_remaining)
        self.rows_processed += self.rows_per_chunk
        estimated_number_of_chunks = (
            i + math.ceil(rows_remaining / self.rows_per_chunk) if rows_remaining else i
        )

        self.history.setdefault(C_TIME, []).append(
            datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S.%f")
        )
        self.history.setdefault(C_DEPTH, []).append(self.depth)
        for m in METRICS:
            self.history.setdefault(f"cum_overhead_{m}", []).append(
                self.cum_overhead[m]
            )
        self.history.setdefault(C_NUM_ROWS, []).append(prev_cum_rows)
        self.history.setdefault("chunk", []).append(i)
        self.history.setdefault("chunk_size", []).append(self.chunk_size)
        self.history.setdefault("row_size", []).append(observed_row_size)

        # diagnostics not reported by ChunkHistorian

        if chunk_metric() == USS:
            self.history.setdefault("prev_uss", []).append(prev_uss)
            self.history.setdefault("cur_uss", []).append(self.uss)
        else:
            self.history.setdefault("prev_rss", []).append(prev_rss)
            self.history.setdefault("cur_rss", []).append(self.rss)

        self.history.setdefault("prev_headroom", []).append(prev_headroom)
        self.history.setdefault("cur_headroom", []).append(self.headroom)

        for m in METRICS:
            self.history.setdefault(f"overhead_{m}", []).append(overhead[m])

        self.history.setdefault("new_rows_processed", []).append(self.rows_processed)
        self.history.setdefault("new_rows_per_chunk", []).append(self.rows_per_chunk)
        self.history.setdefault("estimated_num_chunks", []).append(
            estimated_number_of_chunks
        )

        history_df = pd.DataFrame.from_dict(self.history)
        if LOG_SUBCHUNK_HISTORY:
            logger.debug(
                f"ChunkSizer.adaptive_rows_per_chunk {self.chunk_tag}\n{history_df.transpose()}"
            )

        # input()

        if chunk_training_mode() not in (MODE_PRODUCTION, MODE_CHUNKLESS):
            self.cum_rows += self.rows_per_chunk

        return self.rows_per_chunk, estimated_number_of_chunks

    @contextmanager
    def ledger(self):

        # don't do anything in chunkless mode
        if chunk_training_mode() == MODE_CHUNKLESS:
            yield
            return

        mem_monitor = None

        # nested chunkers should be unchunked
        if len(CHUNK_LEDGERS) > 0:
            assert self.chunk_size == 0

        with ledger_lock:
            self.chunk_ledger = ChunkLedger(
                self.trace_label, self.chunk_size, self.rss, self.uss, self.headroom
            )
            CHUNK_LEDGERS.append(self.chunk_ledger)

        # reality check - there should be one ledger per sizer
        assert len(CHUNK_LEDGERS) == len(CHUNK_SIZERS)

        try:
            # all calls to log_df within this block will be directed to top level chunk_ledger
            # and passed on down the stack to the base to support hwm tallies

            # if this is a base chunk_sizer (and ledger) then start a thread to monitor rss usage
            if (len(CHUNK_LEDGERS) == 1) and ENABLE_MEMORY_MONITOR:
                stop_snooping = threading.Event()
                mem_monitor = MemMonitor(self.trace_label, stop_snooping)
                mem_monitor.start()

            log_rss(
                self.trace_label, force=True
            )  # make sure we get at least one reading
            yield
            log_rss(
                self.trace_label, force=True
            )  # make sure we get at least one reading

        finally:

            if mem_monitor is not None:

                if not mem_monitor.is_alive():
                    logger.error(f"mem_monitor for {self.trace_label} died!")
                    bug  # bug

                stop_snooping.set()
                while mem_monitor.is_alive():
                    logger.debug(
                        f"{self.trace_label} waiting for mem_monitor thread to terminate"
                    )
                    mem_monitor.join(timeout=MEM_MONITOR_TICK)

            with ledger_lock:
                self.chunk_ledger.close()
                CHUNK_LEDGERS.pop()
                self.chunk_ledger = None


@contextmanager
def chunk_log(trace_label, chunk_tag=None, base=False):

    # With `base=True` this method can be used to instantiate
    # a ChunkSizer class object without actually chunking. This
    # avoids breaking the assertion below.

    assert base == (len(CHUNK_SIZERS) == 0)

    trace_label = f"{trace_label}.chunk_log"

    chunk_tag = chunk_tag or trace_label
    num_choosers = 0
    chunk_size = 0

    chunk_sizer = ChunkSizer(chunk_tag, trace_label, num_choosers, chunk_size)

    chunk_sizer.initial_rows_per_chunk()

    with chunk_sizer.ledger():

        yield

        if chunk_training_mode() != MODE_CHUNKLESS:
            chunk_sizer.adaptive_rows_per_chunk(1)

    chunk_sizer.close()


@contextmanager
def chunk_log_skip():

    yield

    None


def adaptive_chunked_choosers(choosers, chunk_size, trace_label, chunk_tag=None):

    # generator to iterate over choosers

    chunk_tag = chunk_tag or trace_label

    num_choosers = len(choosers.index)
    assert num_choosers > 0
    assert chunk_size >= 0

    logger.info(
        f"{trace_label} Running adaptive_chunked_choosers with {num_choosers} choosers"
    )

    chunk_sizer = ChunkSizer(chunk_tag, trace_label, num_choosers, chunk_size)

    rows_per_chunk, estimated_number_of_chunks = chunk_sizer.initial_rows_per_chunk()

    i = offset = 0
    while offset < num_choosers:

        i += 1
        assert offset + rows_per_chunk <= num_choosers

        chunk_trace_label = trace_label_for_chunk(trace_label, chunk_size, i)

        with chunk_sizer.ledger():

            # grab the next chunk based on current rows_per_chunk
            chooser_chunk = choosers[offset : offset + rows_per_chunk]

            logger.info(
                f"Running chunk {i} of {estimated_number_of_chunks or '?'} "
                f"with {len(chooser_chunk)} of {num_choosers} choosers"
            )

            yield i, chooser_chunk, chunk_trace_label

            offset += rows_per_chunk

            if chunk_training_mode() != MODE_CHUNKLESS:
                (
                    rows_per_chunk,
                    estimated_number_of_chunks,
                ) = chunk_sizer.adaptive_rows_per_chunk(i)

    chunk_sizer.close()


def adaptive_chunked_choosers_and_alts(
    choosers, alternatives, chunk_size, trace_label, chunk_tag=None
):
    """
    generator to iterate over choosers and alternatives in chunk_size chunks

    like chunked_choosers, but also chunks alternatives
    for use with sampled alternatives which will have different alternatives (and numbers of alts)

    There may be up to sample_size (or as few as one) alternatives for each chooser
    because alternatives may have been sampled more than once,  but pick_count for those
    alternatives will always sum to sample_size.

    When we chunk the choosers, we need to take care chunking the alternatives as there are
    varying numbers of them for each chooser. Since alternatives appear in the same order
    as choosers, we can use cumulative pick_counts to identify boundaries of sets of alternatives

    Parameters
    ----------
    choosers
    alternatives : pandas DataFrame
        sample alternatives including pick_count column in same order as choosers
    rows_per_chunk : int

    Yields
    -------
    i : int
        one-based index of current chunk
    num_chunks : int
        total number of chunks that will be yielded
    choosers : pandas DataFrame slice
        chunk of choosers
    alternatives : pandas DataFrame slice
        chunk of alternatives for chooser chunk
    """

    chunk_tag = chunk_tag or trace_label

    num_choosers = len(choosers.index)
    num_alternatives = len(alternatives.index)
    assert num_choosers > 0

    # alternatives index should match choosers (except with duplicate repeating alt rows)
    assert choosers.index.equals(
        alternatives.index[~alternatives.index.duplicated(keep="first")]
    )

    last_repeat = alternatives.index != np.roll(alternatives.index, -1)

    assert (num_choosers == 1) or choosers.index.equals(alternatives.index[last_repeat])
    assert (
        "pick_count" in alternatives.columns
        or choosers.index.name == alternatives.index.name
    )
    assert choosers.index.name == alternatives.index.name

    logger.info(
        f"{trace_label} Running adaptive_chunked_choosers_and_alts "
        f"with {num_choosers} choosers and {num_alternatives} alternatives"
    )

    chunk_sizer = ChunkSizer(chunk_tag, trace_label, num_choosers, chunk_size)
    rows_per_chunk, estimated_number_of_chunks = chunk_sizer.initial_rows_per_chunk()
    assert (rows_per_chunk > 0) and (rows_per_chunk <= num_choosers)

    # alt chunks boundaries are where index changes
    alt_ids = alternatives.index.values
    alt_chunk_ends = np.where(alt_ids[:-1] != alt_ids[1:])[0] + 1
    alt_chunk_ends = np.append(
        [0], alt_chunk_ends
    )  # including the first to simplify indexing
    alt_chunk_ends = np.append(
        alt_chunk_ends, [len(alternatives.index)]
    )  # end of final chunk

    i = offset = alt_offset = 0
    while offset < num_choosers:
        i += 1

        assert (
            offset + rows_per_chunk <= num_choosers
        ), f"i {i} offset {offset} rows_per_chunk {rows_per_chunk} num_choosers {num_choosers}"

        chunk_trace_label = trace_label_for_chunk(trace_label, chunk_size, i)

        with chunk_sizer.ledger():

            chooser_chunk = choosers[offset : offset + rows_per_chunk]

            alt_end = alt_chunk_ends[offset + rows_per_chunk]
            alternative_chunk = alternatives[alt_offset:alt_end]

            assert len(chooser_chunk.index) == len(
                np.unique(alternative_chunk.index.values)
            )
            assert (
                chooser_chunk.index == np.unique(alternative_chunk.index.values)
            ).all()

            logger.info(
                f"Running chunk {i} of {estimated_number_of_chunks or '?'} "
                f"with {len(chooser_chunk)} of {num_choosers} choosers"
            )

            yield i, chooser_chunk, alternative_chunk, chunk_trace_label

            offset += rows_per_chunk
            alt_offset = alt_end

            if chunk_training_mode() != MODE_CHUNKLESS:
                (
                    rows_per_chunk,
                    estimated_number_of_chunks,
                ) = chunk_sizer.adaptive_rows_per_chunk(i)

    chunk_sizer.close()


def adaptive_chunked_choosers_by_chunk_id(
    choosers, chunk_size, trace_label, chunk_tag=None
):
    # generator to iterate over choosers in chunk_size chunks
    # like chunked_choosers but based on chunk_id field rather than dataframe length
    # (the presumption is that choosers has multiple rows with the same chunk_id that
    # all have to be included in the same chunk)
    # FIXME - we pathologically know name of chunk_id col in households table

    chunk_tag = chunk_tag or trace_label

    num_choosers = choosers["chunk_id"].max() + 1
    assert num_choosers > 0

    chunk_sizer = ChunkSizer(chunk_tag, trace_label, num_choosers, chunk_size)

    rows_per_chunk, estimated_number_of_chunks = chunk_sizer.initial_rows_per_chunk()

    i = offset = 0
    while offset < num_choosers:

        i += 1
        assert offset + rows_per_chunk <= num_choosers

        chunk_trace_label = trace_label_for_chunk(trace_label, chunk_size, i)

        with chunk_sizer.ledger():

            chooser_chunk = choosers[
                choosers["chunk_id"].between(offset, offset + rows_per_chunk - 1)
            ]

            logger.info(
                f"{trace_label} Running chunk {i} of {estimated_number_of_chunks or '?'} "
                f"with {rows_per_chunk} of {num_choosers} choosers"
            )

            yield i, chooser_chunk, chunk_trace_label

            offset += rows_per_chunk

            if chunk_training_mode() != MODE_CHUNKLESS:
                (
                    rows_per_chunk,
                    estimated_number_of_chunks,
                ) = chunk_sizer.adaptive_rows_per_chunk(i)

    chunk_sizer.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging
import os
import warnings

import pandas as pd

from activitysim.abm.tables import shadow_pricing
from activitysim.core import chunk, config, expressions, inject, mem, pipeline, tracing
from activitysim.core.steps.output import (
    track_skim_usage,
    write_data_dictionary,
    write_tables,
)

# We are using the naming conventions in the mtc_asim.h5 example
# file for our default list. This provides backwards compatibility
# with previous versions of ActivitySim in which only 'input_store'
# is given in the settings file.
DEFAULT_TABLE_LIST = [
    {
        "tablename": "households",
        "h5_tablename": "households",
        "index_col": "household_id",
    },
    {"tablename": "persons", "h5_tablename": "persons", "index_col": "person_id"},
    {"tablename": "land_use", "h5_tablename": "land_use_taz", "index_col": "TAZ"},
]

logger = logging.getLogger(__name__)


def annotate_tables(model_settings, trace_label):

    trace_label = tracing.extend_trace_label(trace_label, "annotate_tables")

    chunk.log_rss(trace_label)

    annotate_tables = model_settings.get("annotate_tables", [])

    if not annotate_tables:
        logger.warning(
            f"{trace_label} - annotate_tables setting is empty - nothing to do!"
        )

    assert isinstance(
        annotate_tables, list
    ), f"annotate_tables settings should be a list but is {type(annotate_tables)}"

    t0 = tracing.print_elapsed_time()

    for table_info in annotate_tables:

        tablename = table_info["tablename"]

        chunk.log_rss(f"{trace_label}.pre-get_table.{tablename}")

        df = inject.get_table(tablename).to_frame()
        chunk.log_df(trace_label, tablename, df)

        # - rename columns
        column_map = table_info.get("column_map", None)
        if column_map:

            warnings.warn(
                f"Setting 'column_map' has been changed to 'rename_columns'. "
                f"Support for 'column_map' in annotate_tables  will be removed in future versions.",
                FutureWarning,
            )

            logger.info(f"{trace_label} - renaming {tablename} columns {column_map}")
            df.rename(columns=column_map, inplace=True)

        # - annotate
        annotate = table_info.get("annotate", None)
        if annotate:
            logger.info(
                f"{trace_label} - annotating {tablename} SPEC {annotate['SPEC']}"
            )
            expressions.assign_columns(
                df=df, model_settings=annotate, trace_label=trace_label
            )

        chunk.log_df(trace_label, tablename, df)

        # - write table to pipeline
        pipeline.replace_table(tablename, df)

        del df
        chunk.log_df(trace_label, tablename, None)


@inject.step()
def initialize_landuse():

    trace_label = "initialize_landuse"

    with chunk.chunk_log(trace_label, base=True):

        model_settings = config.read_model_settings(
            "initialize_landuse.yaml", mandatory=True
        )

        annotate_tables(model_settings, trace_label)

        # instantiate accessibility (must be checkpointed to be be used to slice accessibility)
        accessibility = pipeline.get_table("accessibility")
        chunk.log_df(trace_label, "accessibility", accessibility)


@inject.step()
def initialize_households():

    trace_label = "initialize_households"

    with chunk.chunk_log(trace_label, base=True):

        chunk.log_rss(f"{trace_label}.inside-yield")

        households = inject.get_table("households").to_frame()
        assert not households._is_view
        chunk.log_df(trace_label, "households", households)
        del households
        chunk.log_df(trace_label, "households", None)

        persons = inject.get_table("persons").to_frame()
        assert not persons._is_view
        chunk.log_df(trace_label, "persons", persons)
        del persons
        chunk.log_df(trace_label, "persons", None)

        model_settings = config.read_model_settings(
            "initialize_households.yaml", mandatory=True
        )
        annotate_tables(model_settings, trace_label)

        # - initialize shadow_pricing size tables after annotating household and person tables
        # since these are scaled to model size, they have to be created while single-process
        # this can now be called as a stand alone model step instead, add_size_tables
        add_size_tables = model_settings.get("add_size_tables", True)
        if add_size_tables:
            # warnings.warn(f"Calling add_size_tables from initialize will be removed in the future.", FutureWarning)
            shadow_pricing.add_size_tables()

        # - preload person_windows
        person_windows = inject.get_table("person_windows").to_frame()
        chunk.log_df(trace_label, "person_windows", person_windows)


@inject.injectable(cache=True)
def preload_injectables():
    """
    preload bulky injectables up front - stuff that isn't inserted into the pipeline
    """

    logger.info("preload_injectables")

    inject.add_step("track_skim_usage", track_skim_usage)
    inject.add_step("write_data_dictionary", write_data_dictionary)
    inject.add_step("write_tables", write_tables)

    table_list = config.setting("input_table_list")

    # default ActivitySim table names and indices
    if table_list is None:
        logger.warning(
            "No 'input_table_list' found in settings. This will be a "
            "required setting in upcoming versions of ActivitySim."
        )

        new_settings = inject.get_injectable("settings")
        new_settings["input_table_list"] = DEFAULT_TABLE_LIST
        inject.add_injectable("settings", new_settings)

    # FIXME undocumented feature
    if config.setting("write_raw_tables"):

        # write raw input tables as csv (before annotation)
        csv_dir = config.output_file_path("raw_tables")
        if not os.path.exists(csv_dir):
            os.makedirs(csv_dir)  # make directory if needed

        table_names = [t["tablename"] for t in table_list]
        for t in table_names:
            df = inject.get_table(t).to_frame()
            df.to_csv(os.path.join(csv_dir, "%s.csv" % t), index=True)

    t0 = tracing.print_elapsed_time()

    if config.setting("benchmarking", False):
        # we don't want to pay for skim_dict inside any model component during
        # benchmarking, so we'll preload skim_dict here.  Preloading is not needed
        # for regular operation, as activitysim components can load-on-demand.
        if inject.get_injectable("skim_dict", None) is not None:
            t0 = tracing.print_elapsed_time("preload skim_dict", t0, debug=True)

    return True



# ActivitySim
# See full license in LICENSE.txt.

import os.path

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

from .. import inject, simulate


@pytest.fixture(scope="module")
def data_dir():
    return os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def spec_name(data_dir):
    return "sample_spec.csv"


@pytest.fixture(scope="module")
def spec(data_dir, spec_name):
    return simulate.read_model_spec(file_name=spec_name)


@pytest.fixture(scope="module")
def data(data_dir):
    return pd.read_csv(os.path.join(data_dir, "data.csv"))


def setup_function():
    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)

    output_dir = os.path.join(os.path.dirname(__file__), f"output")
    inject.add_injectable("output_dir", output_dir)


def test_read_model_spec(spec_name):

    spec = simulate.read_model_spec(file_name=spec_name)

    assert len(spec) == 4
    assert spec.index.name == "Expression"
    assert list(spec.columns) == ["alt0", "alt1"]
    npt.assert_array_equal(spec.values, [[1.1, 11], [2.2, 22], [3.3, 33], [4.4, 44]])


def test_eval_variables(spec, data):

    result = simulate.eval_variables(spec.index, data)

    expected = pd.DataFrame(
        [[1, 0, 4, 1], [0, 1, 4, 1], [0, 1, 5, 1]], index=data.index, columns=spec.index
    )

    expected[expected.columns[0]] = expected[expected.columns[0]].astype(np.int8)
    expected[expected.columns[1]] = expected[expected.columns[1]].astype(np.int8)
    expected[expected.columns[2]] = expected[expected.columns[2]].astype(np.int64)
    expected[expected.columns[3]] = expected[expected.columns[3]].astype(int)

    print("\nexpected\n%s" % expected.dtypes)
    print("\nresult\n%s" % result.dtypes)

    pdt.assert_frame_equal(result, expected, check_names=False)


def test_simple_simulate(data, spec):

    inject.add_injectable("settings", {"check_for_variability": False})

    choices = simulate.simple_simulate(choosers=data, spec=spec, nest_spec=None)
    expected = pd.Series([1, 1, 1], index=data.index)
    pdt.assert_series_equal(choices, expected)


def test_simple_simulate_chunked(data, spec):

    inject.add_injectable("settings", {"check_for_variability": False})

    choices = simulate.simple_simulate(
        choosers=data, spec=spec, nest_spec=None, chunk_size=2
    )
    expected = pd.Series([1, 1, 1], index=data.index)
    pdt.assert_series_equal(choices, expected)



import pandas as pd

from activitysim.core import inject, pipeline, tracing


@inject.step()
def step1():

    table1 = pd.DataFrame({"c": [1, 2, 3]})
    inject.add_table("table1", table1)


@inject.step()
def step2():

    table1 = pd.DataFrame({"c": [2, 4, 6]})
    inject.add_table("table2", table1)


@inject.step()
def step3():

    table1 = pd.DataFrame({"c": [3, 6, 9]})
    inject.add_table("table3", table1)


@inject.step()
def step_add_col():

    table_name = inject.get_step_arg("table_name")
    assert table_name is not None

    col_name = inject.get_step_arg("column_name")
    assert col_name is not None

    table = pipeline.get_table(table_name)

    assert col_name not in table.columns

    table[col_name] = table.index + (1000 * len(table.columns))

    pipeline.replace_table(table_name, table)


@inject.step()
def step_forget_tab():

    table_name = inject.get_step_arg("table_name")
    assert table_name is not None

    table = pipeline.get_table(table_name)

    pipeline.drop_table(table_name)


@inject.step()
def create_households(trace_hh_id):

    df = pd.DataFrame({"household_id": [1, 2, 3], "home_zone_id": {100, 100, 101}})
    inject.add_table("households", df)

    pipeline.get_rn_generator().add_channel("households", df)

    tracing.register_traceable_table("households", df)



# ActivitySim
# See full license in LICENSE.txt.
import logging

from orca import orca

_DECORATED_STEPS = {}
_DECORATED_TABLES = {}
_DECORATED_COLUMNS = {}
_DECORATED_INJECTABLES = {}


# we want to allow None (any anyting else) as a default value, so just choose an improbable string
_NO_DEFAULT = "throw error if missing"

logger = logging.getLogger(__name__)


def step():
    def decorator(func):
        name = func.__name__

        logger.debug("inject step %s" % name)

        assert not _DECORATED_STEPS.get(name, False), (
            "step '%s' already decorated." % name
        )
        _DECORATED_STEPS[name] = func

        orca.add_step(name, func)

        return func

    return decorator


def table():
    def decorator(func):
        name = func.__name__

        logger.debug("inject table %s" % name)

        assert not _DECORATED_TABLES.get(name, False), (
            "table '%s' already decorated." % name
        )
        _DECORATED_TABLES[name] = func

        orca.add_table(name, func)

        return func

    return decorator


def injectable(cache=False, override=False):
    def decorator(func):
        name = func.__name__

        logger.debug("inject injectable %s" % name)

        # insist on explicit override to ensure multiple definitions occur in correct order
        assert override or not _DECORATED_INJECTABLES.get(name, False), (
            "injectable '%s' already defined. not overridden" % name
        )

        _DECORATED_INJECTABLES[name] = {"func": func, "cache": cache}

        orca.add_injectable(name, func, cache=cache)

        return func

    return decorator


def merge_tables(target, tables, columns=None):
    return orca.merge_tables(target, tables, columns)


def add_step(name, func):
    return orca.add_step(name, func)


def add_table(table_name, table, replace=False):
    """
    Add new table and raise assertion error if the table already exists.
    Silently replace if replace=True.
    """
    if (
        not replace
        and orca.is_table(table_name)
        and orca.table_type(table_name) == "dataframe"
    ):
        logger.warning("inject add_table replacing existing table %s" % table_name)
        assert False

    # FIXME - should add table.copy() instead, so it can't be modified behind our back?
    return orca.add_table(table_name, table, cache=False)


# fixme remove?
def add_column(table_name, column_name, column, cache=False):
    return orca.add_column(table_name, column_name, column, cache=cache)


def add_injectable(name, injectable, cache=False):
    return orca.add_injectable(name, injectable, cache=cache)


def broadcast(
    cast, onto, cast_on=None, onto_on=None, cast_index=False, onto_index=False
):
    return orca.broadcast(
        cast,
        onto,
        cast_on=cast_on,
        onto_on=onto_on,
        cast_index=cast_index,
        onto_index=onto_index,
    )


def get_table(name, default=_NO_DEFAULT):

    if orca.is_table(name) or default == _NO_DEFAULT:
        return orca.get_table(name)
    else:
        return default


def is_injectable(name):

    return orca.is_injectable(name)


def get_injectable(name, default=_NO_DEFAULT):

    if is_injectable(name) or default == _NO_DEFAULT:
        return orca.get_injectable(name)
    else:
        return default


def remove_injectable(name):

    orca._INJECTABLES.pop(name, None)


def reinject_decorated_tables():
    """
    reinject the decorated tables (and columns)
    """

    logger.info("reinject_decorated_tables")

    # need to clear any non-decorated tables that were added during the previous run
    orca._TABLES.clear()
    orca._COLUMNS.clear()
    orca._TABLE_CACHE.clear()
    orca._COLUMN_CACHE.clear()

    for name, func in _DECORATED_TABLES.items():
        logger.debug("reinject decorated table %s" % name)
        orca.add_table(name, func)

    for column_key, args in _DECORATED_COLUMNS.items():
        table_name, column_name = column_key
        logger.debug("reinject decorated column %s.%s" % (table_name, column_name))
        orca.add_column(table_name, column_name, args["func"], cache=args["cache"])

    for name, args in _DECORATED_INJECTABLES.items():
        logger.debug("reinject decorated injectable %s" % name)
        orca.add_injectable(name, args["func"], cache=args["cache"])


def clear_cache():
    return orca.clear_cache()


def set_step_args(args=None):

    assert isinstance(args, dict) or args is None
    orca.add_injectable("step_args", args)


def get_step_arg(arg_name, default=_NO_DEFAULT):

    args = orca.get_injectable("step_args")

    assert isinstance(args, dict)
    if arg_name not in args and default == _NO_DEFAULT:
        raise "step arg '%s' not found and no default" % arg_name

    return args.get(arg_name, default)


def dump_state():

    print("_DECORATED_STEPS", list(_DECORATED_STEPS.keys()))
    print("orca._STEPS", list(orca._STEPS.keys()))



# ActivitySim
# See full license in LICENSE.txt.
import os

import pytest

from activitysim.core import inject

# The following import statement has the side-effect of registering injectables:
from .. import __init__


def test_misc():

    inject.clear_cache()

    with pytest.raises(RuntimeError) as excinfo:
        inject.get_injectable("configs_dir")
    assert "directory does not exist" in str(excinfo.value)

    with pytest.raises(RuntimeError) as excinfo:
        inject.get_injectable("data_dir")
    assert "directory does not exist" in str(excinfo.value)

    with pytest.raises(RuntimeError) as excinfo:
        inject.get_injectable("output_dir")
    assert "directory does not exist" in str(excinfo.value)

    configs_dir = os.path.join(os.path.dirname(__file__), "configs_test_misc")
    inject.add_injectable("configs_dir", configs_dir)

    settings = inject.get_injectable("settings")
    assert isinstance(settings, dict)

    data_dir = os.path.join(os.path.dirname(__file__), "data")
    inject.add_injectable("data_dir", data_dir)

    # default values if not specified in settings
    assert inject.get_injectable("chunk_size") == 0



# ActivitySim
# See full license in LICENSE.txt.
import os
import subprocess

import pandas as pd
import pandas.testing as pdt
import pkg_resources

from activitysim.core import inject


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def run_test_mtc(multiprocess=False, chunkless=False):
    def example_path(dirname):
        resource = os.path.join("examples", "prototype_mtc", dirname)
        return pkg_resources.resource_filename("activitysim", resource)

    def test_path(dirname):
        return os.path.join(os.path.dirname(__file__), dirname)

    def regress():
        regress_trips_df = pd.read_csv(test_path("regress/final_trips.csv"))
        final_trips_df = pd.read_csv(test_path("output/final_trips.csv"))

        # person_id,household_id,tour_id,primary_purpose,trip_num,outbound,trip_count,purpose,
        # destination,origin,destination_logsum,depart,trip_mode,mode_choice_logsum
        # compare_cols = []
        pdt.assert_frame_equal(final_trips_df, regress_trips_df)

    file_path = os.path.join(os.path.dirname(__file__), "simulation.py")

    if multiprocess:
        run_args = [
            "-c",
            test_path("configs_mp"),
            "-c",
            example_path("configs_mp"),
            "-c",
            example_path("configs"),
            "-d",
            example_path("data"),
            "-o",
            test_path("output"),
        ]
    elif chunkless:
        run_args = [
            "-c",
            test_path("configs_chunkless"),
            "-c",
            example_path("configs"),
            "-d",
            example_path("data"),
            "-o",
            test_path("output"),
        ]
    else:
        run_args = [
            "-c",
            test_path("configs"),
            "-c",
            example_path("configs"),
            "-d",
            example_path("data"),
            "-o",
            test_path("output"),
        ]

    subprocess.run(["coverage", "run", "-a", file_path] + run_args, check=True)

    regress()


def test_mtc():
    run_test_mtc(multiprocess=False)


def test_mtc_chunkless():
    run_test_mtc(multiprocess=False, chunkless=True)


def test_mtc_mp():
    run_test_mtc(multiprocess=True)


if __name__ == "__main__":

    run_test_mtc(multiprocess=False)
    run_test_mtc(multiprocess=True)
    run_test_mtc(multiprocess=False, chunkless=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.abm.tables import shadow_pricing
from activitysim.core import (
    config,
    expressions,
    inject,
    logit,
    los,
    mem,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.interaction_sample import interaction_sample
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate

from .util import estimation
from .util import logsums as logsum
from .util import tour_destination

# import multiprocessing


"""
The school/workplace location model predicts the zones in which various people will
work or attend school.

For locations choices like workplace and school location, we have existing data about the actual
number of workers or students in the various destination zones, and we naturally want the results
of location choice to yield distributions the match these observed distributions as closely as
possible. To achieve this, we use start with size tables with the observed populations by zone
and segment (e.g. number of university, highschool, and gradeschool students in each zone) and
use those populations as attractors (positive utilities) so that high population zones will,
all things being equal, receive more choices. (For instance, we want university-goers to choose
school locations with in zones with university enrollments.)

But since the choice algorithm can result in aggregate distributions of choices (modeled_size)
that don't match observed (predicted_size) counts. The shadow pricing algorithm attempts to
correct these misalignments, by iteratively running the choice model, comparing the modeled_size
of the zones segments to the  predicted size, and computing a shadow_price coefficient that is
applied to the size term to boost or attenuate its influence. This iterative process can be
configures to continue until a specified closeness of fit is achieved, or a maximum number of
iterations has occurred. Since the iterative process can be expensive, a facility is provided
to save the computed shadow prices after every iteration, and to load pre-computed shadow prices
on subsequent runs (warm start) to cut down on runtimes.

Since every individual (always person for now) belongs to at most one segment, each segment
(e.g. 'university', 'highschool' , 'gradeschool' for the 'school' location model) is handled
separately and sequentially withing each shadow-price iteration.

The core algorithm has 3 parts:

Because logsum calculations are expensive, rather than computing logsums for all destination
alternatives, we first build a sample of alternatives using simplified (no-logsum) utilities,
and compute logsums only for that sample, and finally chose from among the sampled alternatives.

* run_location_sample - Build a sample destination alternatives using simplified choice criteria
* run_location_logsums - Compute logsums for travel to those alternatives
* run_location_simulate - Rerun the choice model using the logsums to make a final location choice

With shadow pricing, and iterative treatment of each segment, the structure of the code is:

::

    repeat
        for each segment
            run_location_sample
            run_location_logsums
            run_location_simulate
    until convergence
"""

logger = logging.getLogger(__name__)

# column name of logsum in df returned by run_location_logsums (here because used in more than one place)
ALT_LOGSUM = "mode_choice_logsum"


def write_estimation_specs(estimator, model_settings, settings_file):
    """
    write sample_spec, spec, and coefficients to estimation data bundle

    Parameters
    ----------
    model_settings
    settings_file
    """

    estimator.write_model_settings(model_settings, settings_file)
    # estimator.write_spec(model_settings, tag='SAMPLE_SPEC')
    estimator.write_spec(model_settings, tag="SPEC")
    estimator.write_coefficients(model_settings=model_settings)

    estimator.write_table(
        inject.get_injectable("size_terms"), "size_terms", append=False
    )
    estimator.write_table(
        inject.get_table("land_use").to_frame(), "landuse", append=False
    )


def _location_sample(
    segment_name,
    choosers,
    alternatives,
    skims,
    estimator,
    model_settings,
    alt_dest_col_name,
    chunk_size,
    chunk_tag,
    trace_label,
):
    """
    select a sample of alternative locations.

    Logsum calculations are expensive, so we build a table of persons * all zones
    and then select a sample subset of potential locations

    The sample subset is generated by making multiple choices (<sample_size> number of choices)
    which results in sample containing up to <sample_size> choices for each choose (e.g. person)
    and a pick_count indicating how many times that choice was selected for that chooser.)

    person_id,  dest_zone_id, rand,            pick_count
    23750,      14,           0.565502716034,  4
    23750,      16,           0.711135838871,  6
    ...
    23751,      12,           0.408038878552,  1
    23751,      14,           0.972732479292,  2
    """
    assert not choosers.empty

    logger.info("Running %s with %d persons" % (trace_label, len(choosers.index)))

    sample_size = model_settings["SAMPLE_SIZE"]
    if config.setting("disable_destination_sampling", False) or (
        estimator and estimator.want_unsampled_alternatives
    ):
        # FIXME interaction_sample will return unsampled complete alternatives with probs and pick_count
        logger.info(
            "Estimation mode for %s using unsampled alternatives short_circuit_choices"
            % (trace_label,)
        )
        sample_size = 0

    locals_d = {"skims": skims, "segment_size": segment_name}
    constants = config.get_model_constants(model_settings)
    locals_d.update(constants)

    spec = simulate.spec_for_segment(
        model_settings,
        spec_id="SAMPLE_SPEC",
        segment_name=segment_name,
        estimator=estimator,
    )

    # here since presumably we want this when called for either sample or presample
    log_alt_losers = config.setting("log_alt_losers", False)

    choices = interaction_sample(
        choosers,
        alternatives,
        spec=spec,
        sample_size=sample_size,
        alt_col_name=alt_dest_col_name,
        log_alt_losers=log_alt_losers,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    return choices


def location_sample(
    segment_name,
    persons_merged,
    network_los,
    dest_size_terms,
    estimator,
    model_settings,
    chunk_size,
    chunk_tag,
    trace_label,
):

    # FIXME - MEMORY HACK - only include columns actually used in spec
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    choosers = persons_merged[chooser_columns]

    # create wrapper with keys for this lookup - in this case there is a home_zone_id in the choosers
    # and a zone_id in the alternatives which get merged during interaction
    # (logit.interaction_dataset suffixes duplicate chooser column with '_chooser')
    # the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_default_skim_dict()
    skims = skim_dict.wrap("home_zone_id", "zone_id")

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    choices = _location_sample(
        segment_name,
        choosers,
        dest_size_terms,
        skims,
        estimator,
        model_settings,
        alt_dest_col_name,
        chunk_size,
        chunk_tag,
        trace_label,
    )

    return choices


DEST_TAZ = "dest_TAZ"
HOME_TAZ = "TAZ"
HOME_MAZ = "home_zone_id"
DEST_MAZ = "dest_MAZ"


def aggregate_size_terms(dest_size_terms, network_los):
    #
    # aggregate MAZ_size_terms to TAZ_size_terms
    #

    MAZ_size_terms = dest_size_terms.copy()

    # add crosswalk DEST_TAZ column to MAZ_size_terms
    maz_to_taz = (
        network_los.maz_taz_df[["MAZ", "TAZ"]]
        .set_index("MAZ")
        .sort_values(by="TAZ")
        .TAZ
    )
    MAZ_size_terms[DEST_TAZ] = MAZ_size_terms.index.map(maz_to_taz)

    weighted_average_cols = [
        "shadow_price_size_term_adjustment",
        "shadow_price_utility_adjustment",
    ]
    for c in weighted_average_cols:
        MAZ_size_terms[c] *= MAZ_size_terms["size_term"]  # weighted average

    TAZ_size_terms = MAZ_size_terms.groupby(DEST_TAZ).agg(
        {
            "size_term": "sum",
            "shadow_price_size_term_adjustment": "sum",
            "shadow_price_utility_adjustment": "sum",
        }
    )

    for c in weighted_average_cols:
        TAZ_size_terms[c] /= TAZ_size_terms["size_term"]  # weighted average

    if TAZ_size_terms.isna().any(axis=None):
        logger.warning(
            f"TAZ_size_terms with NAN values\n{TAZ_size_terms[TAZ_size_terms.isna().any(axis=1)]}"
        )
        assert not TAZ_size_terms.isna(axis=None).any()

    # print(f"TAZ_size_terms\n{TAZ_size_terms}")
    #           size_term  shadow_price_size_term_adjustment  shadow_price_utility_adjustment
    # dest_TAZ
    # 2             1.419                                1.0                                0
    # 3            20.511                                1.0                                0
    # 4            19.737                                1.0                                0

    MAZ_size_terms = MAZ_size_terms[[DEST_TAZ, "size_term"]].reset_index(drop=False)
    MAZ_size_terms = MAZ_size_terms.sort_values([DEST_TAZ, "zone_id"]).reset_index(
        drop=True
    )

    # print(f"MAZ_size_terms\n{MAZ_size_terms}")
    #       zone_id  dest_TAZ  size_term
    # 0      106097         2      0.774
    # 1      124251         2      0.258
    # 2      124252         2      0.387
    # 3      106165         3      5.031

    return MAZ_size_terms, TAZ_size_terms


def location_presample(
    segment_name,
    persons_merged,
    network_los,
    dest_size_terms,
    estimator,
    model_settings,
    chunk_size,
    chunk_tag,
    trace_label,
):

    trace_label = tracing.extend_trace_label(trace_label, "presample")

    logger.info(f"{trace_label} location_presample")

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    assert DEST_TAZ != alt_dest_col_name

    MAZ_size_terms, TAZ_size_terms = aggregate_size_terms(dest_size_terms, network_los)

    # convert MAZ zone_id to 'TAZ' in choosers (persons_merged)
    # persons_merged[HOME_TAZ] = persons_merged[HOME_MAZ].map(maz_to_taz)
    assert HOME_MAZ in persons_merged
    assert (
        HOME_TAZ in persons_merged
    )  # 'TAZ' should already be in persons_merged from land_use

    # FIXME - MEMORY HACK - only include columns actually used in spec
    # FIXME we don't actually require that land_use provide a TAZ crosswalk
    # FIXME maybe we should add it for multi-zone (from maz_taz) if missing?
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    chooser_columns = [HOME_TAZ if c == HOME_MAZ else c for c in chooser_columns]
    choosers = persons_merged[chooser_columns]

    # create wrapper with keys for this lookup - in this case there is a HOME_TAZ in the choosers
    # and a DEST_TAZ in the alternatives which get merged during interaction
    # the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_skim_dict("taz")
    skims = skim_dict.wrap(HOME_TAZ, DEST_TAZ)

    taz_sample = _location_sample(
        segment_name,
        choosers,
        TAZ_size_terms,
        skims,
        estimator,
        model_settings,
        DEST_TAZ,
        chunk_size,
        chunk_tag,
        trace_label,
    )

    # print(f"taz_sample\n{taz_sample}")
    #            dest_TAZ      prob  pick_count
    # person_id
    # 55227             7  0.009827           1
    # 55227            10  0.000656           1
    # 55227            18  0.014871           1
    # 55227            20  0.035548           3

    # choose a MAZ for each DEST_TAZ choice, choice probability based on MAZ size_term fraction of TAZ total
    maz_choices = tour_destination.choose_MAZ_for_TAZ(
        taz_sample, MAZ_size_terms, trace_label
    )

    assert DEST_MAZ in maz_choices
    maz_choices = maz_choices.rename(columns={DEST_MAZ: alt_dest_col_name})

    return maz_choices


def run_location_sample(
    segment_name,
    persons_merged,
    network_los,
    dest_size_terms,
    estimator,
    model_settings,
    chunk_size,
    chunk_tag,
    trace_label,
):
    """
    select a sample of alternative locations.

    Logsum calculations are expensive, so we build a table of persons * all zones
    and then select a sample subset of potential locations

    The sample subset is generated by making multiple choices (<sample_size> number of choices)
    which results in sample containing up to <sample_size> choices for each choose (e.g. person)
    and a pick_count indicating how many times that choice was selected for that chooser.)

    person_id,  dest_zone_id, rand,            pick_count
    23750,      14,           0.565502716034,  4
    23750,      16,           0.711135838871,  6
    ...
    23751,      12,           0.408038878552,  1
    23751,      14,           0.972732479292,  2
    """

    logger.debug(
        f"dropping {(~(dest_size_terms.size_term > 0)).sum()} "
        f"of {len(dest_size_terms)} rows where size_term is zero"
    )
    dest_size_terms = dest_size_terms[dest_size_terms.size_term > 0]

    # by default, enable presampling for multizone systems, unless they disable it in settings file
    pre_sample_taz = not (network_los.zone_system == los.ONE_ZONE)
    if pre_sample_taz and not config.setting("want_dest_choice_presampling", True):
        pre_sample_taz = False
        logger.info(
            f"Disabled destination zone presampling for {trace_label} "
            f"because 'want_dest_choice_presampling' setting is False"
        )

    if pre_sample_taz:

        logger.info(
            "Running %s location_presample with %d persons"
            % (trace_label, len(persons_merged))
        )

        choices = location_presample(
            segment_name,
            persons_merged,
            network_los,
            dest_size_terms,
            estimator,
            model_settings,
            chunk_size,
            chunk_tag=f"{chunk_tag}.presample",
            trace_label=trace_label,
        )

    else:

        choices = location_sample(
            segment_name,
            persons_merged,
            network_los,
            dest_size_terms,
            estimator,
            model_settings,
            chunk_size,
            chunk_tag=f"{chunk_tag}.sample",
            trace_label=trace_label,
        )

    return choices


def run_location_logsums(
    segment_name,
    persons_merged_df,
    network_los,
    location_sample_df,
    model_settings,
    chunk_size,
    chunk_tag,
    trace_label,
):
    """
    add logsum column to existing location_sample table

    logsum is calculated by running the mode_choice model for each sample (person, dest_zone_id) pair
    in location_sample, and computing the logsum of all the utilities

    +-----------+--------------+----------------+------------+----------------+
    | PERID     | dest_zone_id | rand           | pick_count | logsum (added) |
    +===========+==============+================+============+================+
    | 23750     |  14          | 0.565502716034 | 4          |  1.85659498857 |
    +-----------+--------------+----------------+------------+----------------+
    + 23750     | 16           | 0.711135838871 | 6          | 1.92315598631  |
    +-----------+--------------+----------------+------------+----------------+
    + ...       |              |                |            |                |
    +-----------+--------------+----------------+------------+----------------+
    | 23751     | 12           | 0.408038878552 | 1          | 2.40612135416  |
    +-----------+--------------+----------------+------------+----------------+
    | 23751     | 14           | 0.972732479292 | 2          |  1.44009018355 |
    +-----------+--------------+----------------+------------+----------------+
    """

    assert not location_sample_df.empty

    logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])

    # FIXME - MEMORY HACK - only include columns actually used in spec
    persons_merged_df = logsum.filter_chooser_columns(
        persons_merged_df, logsum_settings, model_settings
    )

    logger.info(
        "Running %s with %s rows" % (trace_label, len(location_sample_df.index))
    )

    choosers = location_sample_df.join(persons_merged_df, how="left")

    tour_purpose = model_settings["LOGSUM_TOUR_PURPOSE"]
    if isinstance(tour_purpose, dict):
        tour_purpose = tour_purpose[segment_name]

    logsums = logsum.compute_logsums(
        choosers,
        tour_purpose,
        logsum_settings,
        model_settings,
        network_los,
        chunk_size,
        chunk_tag,
        trace_label,
    )

    # "add_column series should have an index matching the table to which it is being added"
    # when the index has duplicates, however, in the special case that the series index exactly
    # matches the table index, then the series value order is preserved
    # logsums now does, since workplace_location_sample was on left side of merge de-dup merge
    location_sample_df[ALT_LOGSUM] = logsums

    return location_sample_df


def run_location_simulate(
    segment_name,
    persons_merged,
    location_sample_df,
    network_los,
    dest_size_terms,
    want_logsums,
    estimator,
    model_settings,
    chunk_size,
    chunk_tag,
    trace_label,
):
    """
    run location model on location_sample annotated with mode_choice logsum
    to select a dest zone from sample alternatives

    Returns
    -------
    choices : pandas.DataFrame indexed by persons_merged_df.index
        choice : location choices (zone ids)
        logsum : float logsum of choice utilities across alternatives

    logsums optional & only returned if DEST_CHOICE_LOGSUM_COLUMN_NAME specified in model_settings
    """
    assert not persons_merged.empty

    # FIXME - MEMORY HACK - only include columns actually used in spec
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    choosers = persons_merged[chooser_columns]

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    # alternatives are pre-sampled and annotated with logsums and pick_count
    # but we have to merge additional alt columns into alt sample list
    alternatives = pd.merge(
        location_sample_df,
        dest_size_terms,
        left_on=alt_dest_col_name,
        right_index=True,
        how="left",
    )

    logger.info("Running %s with %d persons" % (trace_label, len(choosers)))

    # create wrapper with keys for this lookup - in this case there is a home_zone_id in the choosers
    # and a zone_id in the alternatives which get merged during interaction
    # the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_default_skim_dict()
    skims = skim_dict.wrap("home_zone_id", alt_dest_col_name)

    locals_d = {"skims": skims, "segment_size": segment_name}
    constants = config.get_model_constants(model_settings)
    if constants is not None:
        locals_d.update(constants)

    if estimator:
        # write choosers after annotation
        estimator.write_choosers(choosers)
        estimator.set_alt_id(alt_dest_col_name)
        estimator.write_interaction_sample_alternatives(alternatives)

    spec = simulate.spec_for_segment(
        model_settings, spec_id="SPEC", segment_name=segment_name, estimator=estimator
    )

    log_alt_losers = config.setting("log_alt_losers", False)

    choices = interaction_sample_simulate(
        choosers,
        alternatives,
        spec=spec,
        choice_column=alt_dest_col_name,
        log_alt_losers=log_alt_losers,
        want_logsums=want_logsums,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
        trace_choice_name=model_settings["DEST_CHOICE_COLUMN_NAME"],
        estimator=estimator,
    )

    if not want_logsums:
        # for consistency, always return a dataframe with canonical column name
        assert isinstance(choices, pd.Series)
        choices = choices.to_frame("choice")

    assert isinstance(choices, pd.DataFrame)

    return choices


def run_location_choice(
    persons_merged_df,
    network_los,
    shadow_price_calculator,
    want_logsums,
    want_sample_table,
    estimator,
    model_settings,
    chunk_size,
    chunk_tag,
    trace_hh_id,
    trace_label,
):
    """
    Run the three-part location choice algorithm to generate a location choice for each chooser

    Handle the various segments separately and in turn for simplicity of expression files

    Parameters
    ----------
    persons_merged_df : pandas.DataFrame
        persons table merged with households and land_use
    network_los : los.Network_LOS
    shadow_price_calculator : ShadowPriceCalculator
        to get size terms
    want_logsums : boolean
    want_sample_table : boolean
    estimator: Estimator object
    model_settings : dict
    chunk_size : int
    trace_hh_id : int
    trace_label : str

    Returns
    -------
    choices : pandas.DataFrame indexed by persons_merged_df.index
        'choice' : location choices (zone ids)
        'logsum' : float logsum of choice utilities across alternatives

    logsums optional & only returned if DEST_CHOICE_LOGSUM_COLUMN_NAME specified in model_settings
    """

    chooser_segment_column = model_settings["CHOOSER_SEGMENT_COLUMN_NAME"]

    # maps segment names to compact (integer) ids
    segment_ids = model_settings["SEGMENT_IDS"]

    choices_list = []
    sample_list = []
    for segment_name, segment_id in segment_ids.items():

        choosers = persons_merged_df[
            persons_merged_df[chooser_segment_column] == segment_id
        ]

        # size_term and shadow price adjustment - one row per zone
        dest_size_terms = shadow_price_calculator.dest_size_terms(segment_name)

        assert (
            dest_size_terms.index.is_monotonic_increasing
        ), f"shadow_price_calculator.dest_size_terms({segment_name}) not monotonic_increasing"

        if choosers.shape[0] == 0:
            logger.info(f"{trace_label} skipping segment {segment_name}: no choosers")
            continue

        # - location_sample
        location_sample_df = run_location_sample(
            segment_name,
            choosers,
            network_los,
            dest_size_terms,
            estimator,
            model_settings,
            chunk_size,
            chunk_tag,  # run_location_sample will add appropriate suffix for sample or presample
            trace_label=tracing.extend_trace_label(
                trace_label, "sample.%s" % segment_name
            ),
        )

        # - location_logsums
        location_sample_df = run_location_logsums(
            segment_name,
            choosers,
            network_los,
            location_sample_df,
            model_settings,
            chunk_size,
            chunk_tag=f"{chunk_tag}.logsums",
            trace_label=tracing.extend_trace_label(
                trace_label, "logsums.%s" % segment_name
            ),
        )

        # - location_simulate
        choices_df = run_location_simulate(
            segment_name,
            choosers,
            location_sample_df,
            network_los,
            dest_size_terms,
            want_logsums,
            estimator,
            model_settings,
            chunk_size,
            chunk_tag=f"{chunk_tag}.simulate",
            trace_label=tracing.extend_trace_label(
                trace_label, "simulate.%s" % segment_name
            ),
        )

        if estimator:
            if trace_hh_id:
                estimation_trace_label = tracing.extend_trace_label(
                    trace_label, f"estimation.{segment_name}.modeled_choices"
                )
                tracing.trace_df(choices_df, label=estimation_trace_label)

            estimator.write_choices(choices_df.choice)
            choices_df.choice = estimator.get_survey_values(
                choices_df.choice,
                "persons",
                column_names=model_settings["DEST_CHOICE_COLUMN_NAME"],
            )
            estimator.write_override_choices(choices_df.choice)

            if want_logsums:
                # if we override choices, we need to to replace choice logsum with ologsim for override location
                # fortunately, as long as we aren't sampling dest alts, the logsum will be in location_sample_df

                # if we start sampling dest alts, we will need code below to compute override location logsum
                assert estimator.want_unsampled_alternatives

                # merge mode_choice_logsum for the overridden location
                # alt_logsums columns: ['person_id', 'choice', 'logsum']
                alt_dest_col = model_settings["ALT_DEST_COL_NAME"]
                alt_logsums = (
                    location_sample_df[[alt_dest_col, ALT_LOGSUM]]
                    .rename(columns={alt_dest_col: "choice", ALT_LOGSUM: "logsum"})
                    .reset_index()
                )

                # choices_df columns: ['person_id', 'choice']
                choices_df = choices_df[["choice"]].reset_index()

                # choices_df columns: ['person_id', 'choice', 'logsum']
                choices_df = pd.merge(choices_df, alt_logsums, how="left").set_index(
                    "person_id"
                )

                logger.debug(
                    f"{trace_label} segment {segment_name} estimation: override logsums"
                )

            if trace_hh_id:
                estimation_trace_label = tracing.extend_trace_label(
                    trace_label, f"estimation.{segment_name}.survey_choices"
                )
                tracing.trace_df(choices_df, estimation_trace_label)

        choices_list.append(choices_df)

        if want_sample_table:
            # FIXME - sample_table
            location_sample_df.set_index(
                model_settings["ALT_DEST_COL_NAME"], append=True, inplace=True
            )
            sample_list.append(location_sample_df)
        else:
            # del this so we dont hold active reference to it while run_location_sample is creating its replacement
            del location_sample_df

    if len(choices_list) > 0:
        choices_df = pd.concat(choices_list)
    else:
        # this will only happen with small samples (e.g. singleton) with no (e.g.) school segs
        logger.warning("%s no choices", trace_label)
        choices_df = pd.DataFrame(columns=["choice", "logsum"])

    if len(sample_list) > 0:
        save_sample_df = pd.concat(sample_list)
    else:
        # this could happen either with small samples as above, or if no saved sample desired
        save_sample_df = None

    return choices_df, save_sample_df


def iterate_location_choice(
    model_settings,
    persons_merged,
    persons,
    households,
    network_los,
    estimator,
    chunk_size,
    trace_hh_id,
    locutor,
    trace_label,
):
    """
    iterate run_location_choice updating shadow pricing until convergence criteria satisfied
    or max_iterations reached.

    (If use_shadow_pricing not enabled, then just iterate once)

    Parameters
    ----------
    model_settings : dict
    persons_merged : injected table
    persons : injected table
    network_los : los.Network_LOS
    chunk_size : int
    trace_hh_id : int
    locutor : bool
        whether this process is the privileged logger of shadow_pricing when multiprocessing
    trace_label : str

    Returns
    -------
    adds choice column model_settings['DEST_CHOICE_COLUMN_NAME']
    adds logsum column model_settings['DEST_CHOICE_LOGSUM_COLUMN_NAME']- if provided
    adds annotations to persons table
    """

    chunk_tag = trace_label

    # boolean to filter out persons not needing location modeling (e.g. is_worker, is_student)
    chooser_filter_column = model_settings["CHOOSER_FILTER_COLUMN_NAME"]

    dest_choice_column_name = model_settings["DEST_CHOICE_COLUMN_NAME"]
    logsum_column_name = model_settings.get("DEST_CHOICE_LOGSUM_COLUMN_NAME")

    sample_table_name = model_settings.get("DEST_CHOICE_SAMPLE_TABLE_NAME")
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    persons_merged_df = persons_merged.to_frame()

    persons_merged_df = persons_merged_df[persons_merged[chooser_filter_column]]

    persons_merged_df.sort_index(
        inplace=True
    )  # interaction_sample expects chooser index to be monotonic increasing

    # chooser segmentation allows different sets coefficients for e.g. different income_segments or tour_types
    chooser_segment_column = model_settings["CHOOSER_SEGMENT_COLUMN_NAME"]

    assert (
        chooser_segment_column in persons_merged_df
    ), f"CHOOSER_SEGMENT_COLUMN '{chooser_segment_column}' not in persons_merged table."

    spc = shadow_pricing.load_shadow_price_calculator(model_settings)
    max_iterations = spc.max_iterations
    assert not (spc.use_shadow_pricing and estimator)

    logger.debug("%s max_iterations: %s" % (trace_label, max_iterations))

    for iteration in range(1, max_iterations + 1):

        if spc.use_shadow_pricing and iteration > 1:
            spc.update_shadow_prices()

        choices_df, save_sample_df = run_location_choice(
            persons_merged_df,
            network_los,
            shadow_price_calculator=spc,
            want_logsums=logsum_column_name is not None,
            want_sample_table=want_sample_table,
            estimator=estimator,
            model_settings=model_settings,
            chunk_size=chunk_size,
            chunk_tag=chunk_tag,
            trace_hh_id=trace_hh_id,
            trace_label=tracing.extend_trace_label(trace_label, "i%s" % iteration),
        )

        # choices_df is a pandas DataFrame with columns 'choice' and (optionally) 'logsum'
        if choices_df is None:
            break

        spc.set_choices(
            choices=choices_df["choice"],
            segment_ids=persons_merged_df[chooser_segment_column].reindex(
                choices_df.index
            ),
        )

        if locutor:
            spc.write_trace_files(iteration)

        if spc.use_shadow_pricing and spc.check_fit(iteration):
            logging.info(
                "%s converged after iteration %s"
                % (
                    trace_label,
                    iteration,
                )
            )
            break

    # - shadow price table
    if locutor:
        if spc.use_shadow_pricing and "SHADOW_PRICE_TABLE" in model_settings:
            inject.add_table(model_settings["SHADOW_PRICE_TABLE"], spc.shadow_prices)
        if "MODELED_SIZE_TABLE" in model_settings:
            inject.add_table(model_settings["MODELED_SIZE_TABLE"], spc.modeled_size)

    persons_df = persons.to_frame()

    # add the choice values to the dest_choice_column in persons dataframe
    # We only chose school locations for the subset of persons who go to school
    # so we backfill the empty choices with -1 to code as no school location
    # names for location choice and (optional) logsums columns
    NO_DEST_ZONE = -1
    persons_df[dest_choice_column_name] = (
        choices_df["choice"].reindex(persons_df.index).fillna(NO_DEST_ZONE).astype(int)
    )

    # add the dest_choice_logsum column to persons dataframe
    if logsum_column_name:
        persons_df[logsum_column_name] = (
            choices_df["logsum"].reindex(persons_df.index).astype("float")
        )

    if save_sample_df is not None:
        # might be None for tiny samples even if sample_table_name was specified
        assert len(save_sample_df.index.get_level_values(0).unique()) == len(choices_df)
        # lest they try to put school and workplace samples into the same table
        if pipeline.is_table(sample_table_name):
            raise RuntimeError(
                "dest choice sample table %s already exists" % sample_table_name
            )
        pipeline.extend_table(sample_table_name, save_sample_df)

    # - annotate persons table
    if "annotate_persons" in model_settings:
        expressions.assign_columns(
            df=persons_df,
            model_settings=model_settings.get("annotate_persons"),
            trace_label=tracing.extend_trace_label(trace_label, "annotate_persons"),
        )

        pipeline.replace_table("persons", persons_df)

        if trace_hh_id:
            tracing.trace_df(persons_df, label=trace_label, warn_if_empty=True)

    # - annotate households table
    if "annotate_households" in model_settings:

        households_df = households.to_frame()
        expressions.assign_columns(
            df=households_df,
            model_settings=model_settings.get("annotate_households"),
            trace_label=tracing.extend_trace_label(trace_label, "annotate_households"),
        )
        pipeline.replace_table("households", households_df)

        if trace_hh_id:
            tracing.trace_df(households_df, label=trace_label, warn_if_empty=True)

    if logsum_column_name:
        tracing.print_summary(
            logsum_column_name, choices_df["logsum"], value_counts=True
        )

    return persons_df


@inject.step()
def workplace_location(
    persons_merged, persons, households, network_los, chunk_size, trace_hh_id, locutor
):
    """
    workplace location choice model

    iterate_location_choice adds location choice column and annotations to persons table
    """

    trace_label = "workplace_location"
    model_settings = config.read_model_settings("workplace_location.yaml")

    estimator = estimation.manager.begin_estimation("workplace_location")
    if estimator:
        write_estimation_specs(estimator, model_settings, "workplace_location.yaml")

    # FIXME - debugging code to test multiprocessing failure handling
    # process_name = multiprocessing.current_process().name
    # if multiprocessing.current_process().name =='mp_households_0':
    #     raise RuntimeError(f"fake fail {process_name}")

    # disable locutor for benchmarking
    if config.setting("benchmarking", False):
        locutor = False

    iterate_location_choice(
        model_settings,
        persons_merged,
        persons,
        households,
        network_los,
        estimator,
        chunk_size,
        trace_hh_id,
        locutor,
        trace_label,
    )

    if estimator:
        estimator.end_estimation()


@inject.step()
def school_location(
    persons_merged, persons, households, network_los, chunk_size, trace_hh_id, locutor
):
    """
    School location choice model

    iterate_location_choice adds location choice column and annotations to persons table
    """

    trace_label = "school_location"
    model_settings = config.read_model_settings("school_location.yaml")

    estimator = estimation.manager.begin_estimation("school_location")
    if estimator:
        write_estimation_specs(estimator, model_settings, "school_location.yaml")

    # disable locutor for benchmarking
    if config.setting("benchmarking", False):
        locutor = False

    iterate_location_choice(
        model_settings,
        persons_merged,
        persons,
        households,
        network_los,
        estimator,
        chunk_size,
        trace_hh_id,
        locutor,
        trace_label,
    )

    if estimator:
        estimator.end_estimation()



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core.util import reindex

logger = logging.getLogger(__name__)


RANDOM_CHANNELS = [
    "households",
    "persons",
    "tours",
    "joint_tour_participants",
    "trips",
    "vehicles",
]
TRACEABLE_TABLES = [
    "households",
    "persons",
    "tours",
    "joint_tour_participants",
    "trips",
    "vehicles",
]

CANONICAL_TABLE_INDEX_NAMES = {
    "households": "household_id",
    "persons": "person_id",
    "tours": "tour_id",
    "joint_tour_participants": "participant_id",
    "trips": "trip_id",
    "land_use": "zone_id",
    "vehicles": "vehicle_id",
}

# unfortunately the two places this is needed (joint_tour_participation and estimation.infer
# don't have much in common in terms of data structures
# candidates['participant_id'] = (candidates[joint_tours.index.name] * MAX_PARTICIPANT_PNUM) + candidates.PNUM
MAX_PARTICIPANT_PNUM = 100


def enumerate_tour_types(tour_flavors):
    # tour_flavors: {'eat': 1, 'business': 2, 'maint': 1}
    # channels:      ['eat1', 'business1', 'business2', 'maint1']
    channels = [
        tour_type + str(tour_num)
        for tour_type, max_count in tour_flavors.items()
        for tour_num in range(1, max_count + 1)
    ]
    return channels


def canonical_tours():
    """
        create labels for every the possible tour by combining tour_type/tour_num.

    Returns
    -------
        list of canonical tour labels in alphabetical order
    """

    # FIXME we pathalogically know what the possible tour_types and their max tour_nums are
    # FIXME instead, should get flavors from alts tables (but we would have to know their names...)
    # alts = pipeline.get_table('non_mandatory_tour_frequency_alts')
    # non_mandatory_tour_flavors = {c : alts[c].max() for c in alts.columns}

    # - non_mandatory_channels
    MAX_EXTENSION = 2
    non_mandatory_tour_flavors = {
        "escort": 2 + MAX_EXTENSION,
        "shopping": 1 + MAX_EXTENSION,
        "othmaint": 1 + MAX_EXTENSION,
        "othdiscr": 1 + MAX_EXTENSION,
        "eatout": 1 + MAX_EXTENSION,
        "social": 1 + MAX_EXTENSION,
    }
    non_mandatory_channels = enumerate_tour_types(non_mandatory_tour_flavors)

    # - mandatory_channels
    mandatory_tour_flavors = {"work": 2, "school": 2}
    mandatory_channels = enumerate_tour_types(mandatory_tour_flavors)

    # - atwork_subtour_channels
    # we need to distinguish between subtours of different work tours
    # (e.g. eat1_1 is eat subtour for parent work tour 1 and eat1_2 is for work tour 2)
    atwork_subtour_flavors = {"eat": 1, "business": 2, "maint": 1}
    atwork_subtour_channels = enumerate_tour_types(atwork_subtour_flavors)
    max_work_tours = mandatory_tour_flavors["work"]
    atwork_subtour_channels = [
        "%s_%s" % (c, i + 1)
        for c in atwork_subtour_channels
        for i in range(max_work_tours)
    ]

    # - joint_tour_channels
    joint_tour_flavors = {
        "shopping": 2,
        "othmaint": 2,
        "othdiscr": 2,
        "eatout": 2,
        "social": 2,
    }
    joint_tour_channels = enumerate_tour_types(joint_tour_flavors)
    joint_tour_channels = ["j_%s" % c for c in joint_tour_channels]

    sub_channels = (
        non_mandatory_channels
        + mandatory_channels
        + atwork_subtour_channels
        + joint_tour_channels
    )

    sub_channels.sort()

    return sub_channels


def set_tour_index(tours, parent_tour_num_col=None, is_joint=False):
    """
    The new index values are stable based on the person_id, tour_type, and tour_num.
    The existing index is ignored and replaced.

    This gives us a stable (predictable) tour_id with tours in canonical order
    (when tours are sorted by tour_id, tours for each person
    of the same type will be adjacent and in increasing tour_type_num order)

    It also simplifies attaching random number streams to tours that are stable
    (even across simulations)

    Parameters
    ----------
    tours : DataFrame
        Tours dataframe to reindex.
    """

    tour_num_col = "tour_type_num"
    possible_tours = canonical_tours()
    possible_tours_count = len(possible_tours)

    assert tour_num_col in tours.columns

    # create string tour_id corresonding to keys in possible_tours (e.g. 'work1', 'j_shopping2')
    tours["tour_id"] = tours.tour_type + tours[tour_num_col].map(str)

    if parent_tour_num_col:
        # we need to distinguish between subtours of different work tours
        # (e.g. eat1_1 is eat subtour for parent work tour 1 and eat1_2 is for work tour 2)

        parent_tour_num = tours[parent_tour_num_col]
        if parent_tour_num.dtype != "int64":
            # might get converted to float if non-subtours rows are None (but we try to avoid this)
            logger.error("parent_tour_num.dtype: %s" % parent_tour_num.dtype)
            parent_tour_num = parent_tour_num.astype(np.int64)

        tours["tour_id"] = tours["tour_id"] + "_" + parent_tour_num.map(str)

    if is_joint:
        tours["tour_id"] = "j_" + tours["tour_id"]

    # map recognized strings to ints
    tours.tour_id = tours.tour_id.replace(
        to_replace=possible_tours, value=list(range(possible_tours_count))
    )

    # convert to numeric - shouldn't be any NaNs - this will raise error if there are
    tours.tour_id = pd.to_numeric(tours.tour_id, errors="raise").astype(np.int64)

    tours.tour_id = (tours.person_id * possible_tours_count) + tours.tour_id

    # if tours.tour_id.duplicated().any():
    #     print("\ntours.tour_id not unique\n%s" % tours[tours.tour_id.duplicated(keep=False)])
    #     print(tours[tours.tour_id.duplicated(keep=False)][['survey_tour_id', 'tour_type', 'tour_category']])
    assert not tours.tour_id.duplicated().any()

    tours.set_index("tour_id", inplace=True, verify_integrity=True)

    # we modify tours in place, but return the dataframe for the convenience of the caller
    return tours


def set_trip_index(trips, tour_id_column="tour_id"):

    MAX_TRIPS_PER_LEG = 4  # max number of trips per leg (inbound or outbound) of tour

    # canonical_trip_num: 1st trip out = 1, 2nd trip out = 2, 1st in = 5, etc.
    canonical_trip_num = (~trips.outbound * MAX_TRIPS_PER_LEG) + trips.trip_num
    trips["trip_id"] = (
        trips[tour_id_column] * (2 * MAX_TRIPS_PER_LEG) + canonical_trip_num
    )
    trips.set_index("trip_id", inplace=True, verify_integrity=True)

    # we modify trips in place, but return the dataframe for the convenience of the caller
    return trips



# ActivitySim
# See full license in LICENSE.txt.

import datetime
import gc
import glob
import logging
import multiprocessing
import os
import platform
import threading
import time

import numpy as np
import pandas as pd
import psutil

from activitysim.core import config, inject, util

logger = logging.getLogger(__name__)

USS = True

GLOBAL_HWM = {}  # to avoid confusion with chunk local hwm

MEM_TRACE_TICK_LEN = 5
MEM_PARENT_TRACE_TICK_LEN = 15
MEM_SNOOP_TICK_LEN = 5
MEM_TICK = 0

MEM_LOG_FILE_NAME = "mem.csv"
OMNIBUS_LOG_FILE_NAME = f"omnibus_mem.csv"

SUMMARY_BIN_SIZE_IN_SECONDS = 15

mem_log_lock = threading.Lock()


def time_bin(timestamps):
    bins_size_in_seconds = SUMMARY_BIN_SIZE_IN_SECONDS
    epoch = pd.Timestamp("1970-01-01")
    seconds_since_epoch = (timestamps - epoch) // pd.Timedelta("1s")
    bin = seconds_since_epoch - (seconds_since_epoch % bins_size_in_seconds)

    return pd.to_datetime(bin, unit="s", origin="unix")


def consolidate_logs():
    """
    Consolidate and aggregate subprocess mem logs
    """

    if not config.setting("multiprocess", False):
        return

    delete_originals = not config.setting("keep_mem_logs", False)
    omnibus_df = []

    # for each multiprocess step
    multiprocess_steps = config.setting("multiprocess_steps", [])
    for step in multiprocess_steps:
        step_name = step.get("name", None)

        logger.debug(f"mem.consolidate_logs for step {step_name}")

        glob_file_name = config.log_file_path(
            f"{step_name}*{MEM_LOG_FILE_NAME}", prefix=False
        )
        glob_files = glob.glob(glob_file_name)

        if not glob_files:
            continue

        logger.debug(
            f"mem.consolidate_logs consolidating {len(glob_files)} logs for {step_name}"
        )

        # for each individual log
        step_summary_df = []
        for f in glob_files:
            df = pd.read_csv(f, comment="#")

            df = df[["rss", "uss", "event", "time"]]

            df.rss = df.rss.astype(np.int64)
            df.uss = df.uss.astype(np.int64)

            df["time"] = time_bin(
                pd.to_datetime(df.time, errors="coerce", format="%Y/%m/%d %H:%M:%S")
            )

            # consolidate events (duplicate rows should be idle steps (e.g. log_rss)
            df = (
                df.groupby("time")
                .agg(
                    rss=("rss", "max"),
                    uss=("uss", "max"),
                )
                .reset_index(drop=False)
            )

            step_summary_df.append(df)  # add step_df to step summary

        # aggregate the individual the logs into a single step log
        step_summary_df = pd.concat(step_summary_df)
        step_summary_df = (
            step_summary_df.groupby("time")
            .agg(rss=("rss", "sum"), uss=("uss", "sum"), num_files=("rss", "size"))
            .reset_index(drop=False)
        )
        step_summary_df = step_summary_df.sort_values("time")

        step_summary_df["step"] = step_name

        # scale missing values (might be missing idle steps for some chunk_tags)
        scale = (
            1
            + (len(glob_files) - step_summary_df.num_files) / step_summary_df.num_files
        )
        for c in ["rss", "uss"]:
            step_summary_df[c] = (step_summary_df[c] * scale).astype(np.int64)

        step_summary_df["scale"] = scale
        del step_summary_df["num_files"]  # do we want to keep track of scale factor?

        if delete_originals:
            util.delete_files(glob_files, f"mem.consolidate_logs.{step_name}")

        # write aggregate step log
        output_path = config.log_file_path(f"mem_{step_name}.csv", prefix=False)
        logger.debug(
            f"chunk.consolidate_logs writing step summary log for step {step_name} to {output_path}"
        )
        step_summary_df.to_csv(output_path, mode="w", index=False)

        omnibus_df.append(step_summary_df)  # add step summary to omnibus

    # aggregate the step logs into a single omnibus log ordered by timestamp
    omnibus_df = pd.concat(omnibus_df)
    omnibus_df = omnibus_df.sort_values("time")

    output_path = config.log_file_path(OMNIBUS_LOG_FILE_NAME, prefix=False)
    logger.debug(f"chunk.consolidate_logs writing omnibus log to {output_path}")
    omnibus_df.to_csv(output_path, mode="w", index=False)


def check_global_hwm(tag, value, label):

    assert value is not None

    hwm = GLOBAL_HWM.setdefault(tag, {})

    is_new_hwm = value > hwm.get("mark", 0) or not hwm
    if is_new_hwm:
        timestamp = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")

        hwm["mark"] = value
        hwm["timestamp"] = timestamp
        hwm["label"] = label

    return is_new_hwm


def log_global_hwm():

    process_name = multiprocessing.current_process().name

    for tag in GLOBAL_HWM:
        hwm = GLOBAL_HWM[tag]
        value = hwm.get("mark", 0)
        logger.info(
            f"{process_name} high water mark {tag}: {util.INT(value)} ({util.GB(value)}) "
            f"timestamp: {hwm.get('timestamp', '<none>')} label:{hwm.get('label', '<none>')}"
        )


def trace_memory_info(event, trace_ticks=0):

    global MEM_TICK

    tick = time.time()
    if trace_ticks and (tick - MEM_TICK < trace_ticks):
        return
    MEM_TICK = tick

    process_name = multiprocessing.current_process().name
    pid = os.getpid()

    current_process = psutil.Process()

    if USS:
        info = current_process.memory_full_info()
        uss = info.uss
    else:
        info = current_process.memory_info()
        uss = 0

    full_rss = rss = info.rss

    num_children = 0
    for child in current_process.children(recursive=True):
        try:
            child_info = child.memory_info()
            full_rss += child_info.rss
            num_children += 1
        except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
            pass

    noteworthy = (
        True  # any reason not to always log this if we are filtering idle ticks?
    )

    noteworthy = (num_children > 0) or noteworthy
    noteworthy = check_global_hwm("rss", full_rss or rss, event) or noteworthy
    noteworthy = check_global_hwm("uss", uss, event) or noteworthy

    if noteworthy:

        # logger.debug(f"trace_memory_info {event} "
        #              f"rss: {GB(full_rss) if num_children else GB(rss)} "
        #              f"uss: {GB(rss)} ")

        timestamp = datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f")  # sortable

        with mem_log_lock:
            MEM_LOG_HEADER = "process,pid,rss,full_rss,uss,event,children,time"
            with config.open_log_file(
                MEM_LOG_FILE_NAME, "a", header=MEM_LOG_HEADER, prefix=True
            ) as log_file:
                print(
                    f"{process_name},"
                    f"{pid},"
                    f"{util.INT(rss)},"  # want these as ints so we can plot them...
                    f"{util.INT(full_rss)},"
                    f"{util.INT(uss)},"
                    f"{event},"
                    f"{num_children},"
                    f"{timestamp}",
                    file=log_file,
                )

    # return rss and uss for optional use by interested callers
    return full_rss or rss, uss


def get_rss(force_garbage_collect=False, uss=False):

    if force_garbage_collect:
        was_disabled = not gc.isenabled()
        if was_disabled:
            gc.enable()
        gc.collect()
        if was_disabled:
            gc.disable()

    if uss:
        info = psutil.Process().memory_full_info()
        return info.rss, info.uss
    else:
        info = psutil.Process().memory_info()
        return info.rss, 0


def shared_memory_size(data_buffers=None):
    """
    return total size of the multiprocessing shared memory block in data_buffers

    Returns
    -------

    """

    shared_size = 0

    if data_buffers is None:
        data_buffers = inject.get_injectable("data_buffers", {})

    for k, data_buffer in data_buffers.items():
        try:
            obj = data_buffer.get_obj()
        except Exception:
            obj = data_buffer
        data = np.ctypeslib.as_array(obj)
        data_size = data.nbytes

        shared_size += data_size

    return shared_size



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.abm.models.util.canonical_ids import set_tour_index
from activitysim.core.util import reindex

logger = logging.getLogger(__name__)


def create_tours(tour_counts, tour_category, parent_col="person_id"):
    """
    This method processes the tour_frequency column that comes
    out of the model of the same name and turns into a DataFrame that
    represents the tours that were generated

    Parameters
    ----------
    tour_counts: DataFrame
        table specifying how many tours of each type to create
        one row per person (or parent_tour for atwork subtours)
        one (int) column per tour_type, with number of tours to create
    tour_category : str
        one of 'mandatory', 'non_mandatory', 'atwork', or 'joint'

    Returns
    -------
    tours : pandas.DataFrame
        An example of a tours DataFrame is supplied as a comment in the
        source code - it has an index which is a unique tour identifier,
        a person_id column, and a tour type column which comes from the
        column names of the alternatives DataFrame supplied above.

        tours.tour_type       - tour type (e.g. school, work, shopping, eat)
        tours.tour_type_num   - if there are two 'school' type tours, they will be numbered 1 and 2
        tours.tour_type_count - number of tours of tour_type parent has (parent's max tour_type_num)
        tours.tour_num        - index of tour (of any type) for parent
        tours.tour_count      - number of tours of any type) for parent (parent's max tour_num)
        tours.tour_category   - one of 'mandatory', 'non_mandatory', 'atwork', or 'joint'
    """

    # FIXME - document requirement to ensure adjacent tour_type_nums in tour_num order

    """
               alt1       alt2     alt3
    <parent_col>
    2588676       2         0         0
    2588677       1         1         0
    """

    # reformat with the columns given below
    tours = tour_counts.stack().reset_index()
    tours.columns = [parent_col, "tour_type", "tour_type_count"]

    """
        <parent_col> tour_type  tour_type_count
    0     2588676    alt1           2
    1     2588676    alt2           0
    2     2588676    alt3           0
    3     2588676    alt1           1
    4     2588677    alt2           1
    5     2588677    alt3           0

    parent_col is the index from non_mandatory_tour_frequency
    tour_type is the column name from non_mandatory_tour_frequency_alts
    tour_type_count is the count value of the tour's chosen alt's tour_type from alts table
    """

    # now do a repeat and a take, so if you have two trips of given type you
    # now have two rows, and zero trips yields zero rows
    tours = tours.take(np.repeat(tours.index.values, tours.tour_type_count.values))

    grouped = tours.groupby([parent_col, "tour_type"])
    tours["tour_type_num"] = grouped.cumcount() + 1
    tours["tour_type_count"] = tours["tour_type_num"] + grouped.cumcount(
        ascending=False
    )

    grouped = tours.groupby(parent_col)
    tours["tour_num"] = grouped.cumcount() + 1
    tours["tour_count"] = tours["tour_num"] + grouped.cumcount(ascending=False)

    """
        <parent_col> tour_type  tour_type_num  tour_type_count tour_num  tour_count
    0     2588676       alt1           1           2               1         4
    0     2588676       alt1           2           2               2         4
    0     2588676       alt2           1           1               3         4
    0     2588676       alt3           1           1               4         4
    """

    # set these here to ensure consistency across different tour categories
    assert tour_category in ["mandatory", "non_mandatory", "atwork", "joint"]
    tours["tour_category"] = tour_category

    # for joint tours, the correct number will be filled in after participation step
    tours["number_of_participants"] = 1

    # index is arbitrary but don't want any duplicates in index
    tours.reset_index(drop=True, inplace=True)

    return tours


def process_tours(
    tour_frequency, tour_frequency_alts, tour_category, parent_col="person_id"
):
    """
    This method processes the tour_frequency column that comes
    out of the model of the same name and turns into a DataFrame that
    represents the tours that were generated

    Parameters
    ----------
    tour_frequency: Series
        A series which has <parent_col> as the index and the chosen alternative
        index as the value
    tour_frequency_alts: DataFrame
        A DataFrame which has as a unique index which relates to the values
        in the series above typically includes columns which are named for trip
        purposes with values which are counts for that trip purpose.  Example
        trip purposes include escort, shopping, othmaint, othdiscr, eatout,
        social, etc.  A row would be an alternative which might be to take
        one shopping trip and zero trips of other purposes, etc.
    tour_category : str
        one of 'mandatory', 'non_mandatory', 'atwork', or 'joint'
    parent_col: str
        the name of the index (parent_tour_id for atwork subtours, otherwise person_id)

    Returns
    -------
    tours : pandas.DataFrame
        An example of a tours DataFrame is supplied as a comment in the
        source code - it has an index which is a unique tour identifier,
        a person_id column, and a tour type column which comes from the
        column names of the alternatives DataFrame supplied above.

        tours.tour_type       - tour type (e.g. school, work, shopping, eat)
        tours.tour_type_num   - if there are two 'school' type tours, they will be numbered 1 and 2
        tours.tour_type_count - number of tours of tour_type parent has (parent's max tour_type_num)
        tours.tour_num        - index of tour (of any type) for parent
        tours.tour_count      - number of tours of any type) for parent (parent's max tour_num)
        tours.tour_category   - one of 'mandatory', 'non_mandatory', 'atwork', or 'joint'
    """

    # FIXME - document requirement to ensure adjacent tour_type_nums in tour_num order

    # get the actual alternatives for each person - have to go back to the
    # non_mandatory_tour_frequency_alts dataframe to get this - the choice
    # above just stored the index values for the chosen alts
    tour_counts = tour_frequency_alts.loc[tour_frequency]

    # assign person ids to the index
    tour_counts.index = tour_frequency.index

    """
               alt1       alt2     alt3
    <parent_col>
    2588676       2         0         0
    2588677       1         1         0
    """

    tours = create_tours(tour_counts, tour_category, parent_col)

    return tours


def process_mandatory_tours(persons, mandatory_tour_frequency_alts):
    """
    This method processes the mandatory_tour_frequency column that comes out of
    the model of the same name and turns into a DataFrame that represents the
    mandatory tours that were generated

    Parameters
    ----------
    persons : DataFrame
        Persons is a DataFrame which has a column call
        mandatory_tour_frequency (which came out of the mandatory tour
        frequency model) and a column is_worker which indicates the person's
        worker status.  The only valid values of the mandatory_tour_frequency
        column to take are "work1", "work2", "school1", "school2" and
        "work_and_school"

    Returns
    -------
    tours : DataFrame
        An example of a tours DataFrame is supplied as a comment in the
        source code - it has an index which is a tour identifier, a person_id
        column, a tour_type column which is "work" or "school" and a tour_num
        column which is set to 1 or 2 depending whether it is the first or
        second mandatory tour made by the person.  The logic for whether the
        work or school tour comes first given a "work_and_school" choice
        depends on the is_worker column: work tours first for workers, second for non-workers
    """

    person_columns = [
        "mandatory_tour_frequency",
        "is_worker",
        "school_zone_id",
        "workplace_zone_id",
        "home_zone_id",
        "household_id",
    ]
    assert not persons.mandatory_tour_frequency.isnull().any()

    tours = process_tours(
        persons.mandatory_tour_frequency.dropna(),
        mandatory_tour_frequency_alts,
        tour_category="mandatory",
    )

    tours_merged = pd.merge(
        tours[["person_id", "tour_type"]],
        persons[person_columns],
        left_on="person_id",
        right_index=True,
    )

    # by default work tours are first for work_and_school tours
    # swap tour_nums for non-workers so school tour is 1 and work is 2
    work_and_school_and_student = (
        tours_merged.mandatory_tour_frequency == "work_and_school"
    ) & ~tours_merged.is_worker

    tours.tour_num = tours.tour_num.where(
        ~work_and_school_and_student, 3 - tours.tour_num
    )

    # work tours destination is workplace_zone_id, school tours destination is school_zone_id
    tours["destination"] = tours_merged.workplace_zone_id.where(
        (tours_merged.tour_type == "work"), tours_merged.school_zone_id
    )

    tours["origin"] = tours_merged.home_zone_id

    tours["household_id"] = tours_merged.household_id

    # assign stable (predictable) tour_id
    set_tour_index(tours)

    """
               person_id tour_type  tour_type_count  tour_type_num  tour_num  tour_count
    tour_id
    12413245      827549    school                2              1         1           2
    12413244      827549    school                2              2         2           2
    12413264      827550      work                1              1         1           2
    12413266      827550    school                1              1         2           2
    ...
               tour_category  destination   household_id
                   mandatory          102         103992
                   mandatory          102         103992
                   mandatory            9         103992
                   mandatory          102         103992
    """
    return tours


def process_non_mandatory_tours(persons, tour_counts):
    """
    This method processes the non_mandatory_tour_frequency column that comes
    out of the model of the same name and turns into a DataFrame that
    represents the non mandatory tours that were generated

    Parameters
    ----------
    persons: pandas.DataFrame
        persons table containing a non_mandatory_tour_frequency column
        which has the index of the chosen alternative as the value
    non_mandatory_tour_frequency_alts: DataFrame
        A DataFrame which has as a unique index which relates to the values
        in the series above typically includes columns which are named for trip
        purposes with values which are counts for that trip purpose.  Example
        trip purposes include escort, shopping, othmaint, othdiscr, eatout,
        social, etc.  A row would be an alternative which might be to take
        one shopping trip and zero trips of other purposes, etc.

    Returns
    -------
    tours : DataFrame
        An example of a tours DataFrame is supplied as a comment in the
        source code - it has an index which is a unique tour identifier,
        a person_id column, and a tour type column which comes from the
        column names of the alternatives DataFrame supplied above.
    """

    tours = create_tours(tour_counts, tour_category="non_mandatory")

    tours["household_id"] = reindex(persons.household_id, tours.person_id)
    tours["origin"] = reindex(persons.home_zone_id, tours.person_id)

    # assign stable (predictable) tour_id
    set_tour_index(tours)

    """
               person_id tour_type  tour_type_count  tour_type_num  tour_num   tour_count
    tour_id
    17008286     1133885  shopping                1              1         1            3
    17008283     1133885  othmaint                1              1         2            3
    17008282     1133885  othdiscr                1              1         3            3
    ...
               tour_category

               non_mandatory
               non_mandatory
               non_mandatory
    """

    return tours


def process_atwork_subtours(work_tours, atwork_subtour_frequency_alts):

    """
    This method processes the atwork_subtour_frequency column that comes
    out of the model of the same name and turns into a DataFrame that
    represents the subtours tours that were generated

    Parameters
    ----------
    work_tours: DataFrame
        A series which has parent work tour tour_id as the index and
        columns with person_id and atwork_subtour_frequency.
    atwork_subtour_frequency_alts: DataFrame
        A DataFrame which has as a unique index with atwork_subtour_frequency values
        and frequency counts for the subtours to be generated for that choice

    Returns
    -------
    tours : DataFrame
        An example of a tours DataFrame is supplied as a comment in the
        source code - it has an index which is a unique tour identifier,
        a person_id column, and a tour type column which comes from the
        column names of the alternatives DataFrame supplied above.
    """

    # print atwork_subtour_frequency_alts
    """
                  eat  business  maint
    alt
    no_subtours     0         0      0
    eat             1         0      0
    business1       0         1      0
    maint           0         0      1
    business2       0         2      0
    eat_business    1         1      0
    """

    parent_col = "parent_tour_id"
    tours = process_tours(
        work_tours.atwork_subtour_frequency.dropna(),
        atwork_subtour_frequency_alts,
        tour_category="atwork",
        parent_col=parent_col,
    )

    # print tours
    """
               parent_tour_id tour_type  tour_type_count  tour_type_num  tour_num  tour_count
    tour_id
    77147972         77147984       eat                1              1         1           2
    77401056         77147984     maint                1              1         2           2
    80893007         80893019       eat                1              1         1           1

              tour_category
                     atwork
                     atwork
                     atwork
    """

    # merge fields from parent work_tours (note parent tour destination becomes subtour origin)
    work_tours = work_tours[["person_id", "household_id", "tour_num", "destination"]]
    work_tours.rename(
        columns={"tour_num": "parent_tour_num", "destination": "origin"}, inplace=True
    )
    tours = pd.merge(tours, work_tours, left_on=parent_col, right_index=True)

    # assign stable (predictable) tour_id
    set_tour_index(tours, parent_tour_num_col="parent_tour_num")

    """
               person_id tour_type  tour_type_count  tour_type_num  tour_num  tour_count
    tour_id
    77147972     5143198       eat                1              1         1           2
    77401056     5143198     maint                1              1         2           2
    80893007     5392867       eat                1              1         1           1

              tour_category   parent_tour_id  household_id
                     atwork         77147984       1625435
                     atwork         77147984       1625435
                     atwork         80893019       2135503
    """

    # don't need this once we have computed index
    del tours["parent_tour_num"]

    return tours


def process_joint_tours(joint_tour_frequency, joint_tour_frequency_alts, point_persons):
    """
    This method processes the joint_tour_frequency column that comes out of
    the model of the same name and turns into a DataFrame that represents the
    joint tours that were generated

    Parameters
    ----------
    joint_tour_frequency : pandas.Series
        household joint_tour_frequency (which came out of the joint tour frequency model)
        indexed by household_id
    joint_tour_frequency_alts: DataFrame
        A DataFrame which has as a unique index with joint_tour_frequency values
        and frequency counts for the tours to be generated for that choice
    point_persons : pandas DataFrame
        table with columns for (at least) person_ids and home_zone_id indexed by household_id

    Returns
    -------
    tours : DataFrame
        An example of a tours DataFrame is supplied as a comment in the
        source code - it has an index which is a tour identifier, a household_id
        column, a tour_type column and tour_type_num and tour_num columns
        which is set to 1 or 2 depending whether it is the first or second joint tour
        made by the household.
    """

    assert not joint_tour_frequency.isnull().any()

    tours = process_tours(
        joint_tour_frequency.dropna(),
        joint_tour_frequency_alts,
        tour_category="joint",
        parent_col="household_id",
    )

    assert not tours.index.duplicated().any()
    assert point_persons.index.name == "household_id"

    # - assign a temp point person to tour so we can create stable index
    tours["person_id"] = reindex(point_persons.person_id, tours.household_id)
    tours["origin"] = reindex(point_persons.home_zone_id, tours.household_id)

    # assign stable (predictable) tour_id
    set_tour_index(tours, is_joint=True)

    """
                   household_id tour_type  tour_type_count  tour_type_num  tour_num  tour_count
    tour_id
    3209530              320953      disc                1              1         1           2
    3209531              320953      disc                2              2         2           2
    23267026            2326702      shop                1              1         1           1
    17978574            1797857      main                1              1         1           1

                   tour_category  tour_category_id  person_id
    3209530                joint                 4     577234
    3209531                joint                 4     577234
    23267026               joint                 4    1742708
    17978574               joint                 4    5143198
    """
    return tours



# ActivitySim
# See full license in LICENSE.txt.
import os

import pandas as pd
import pandas.testing as pdt
from test_multi_zone import example_path, regress_3_zone, setup_dirs

from activitysim.core import inject, mp_tasks, pipeline

# set the max households for all tests (this is to limit memory use on travis)
HOUSEHOLDS_SAMPLE_SIZE = 100


def test_mp_run():

    configs_dir = [example_path("configs_3_zone"), example_path("configs")]
    data_dir = example_path("data_3")

    setup_dirs(configs_dir, data_dir)
    inject.add_injectable("settings_file_name", "settings_mp.yaml")

    run_list = mp_tasks.get_run_list()
    mp_tasks.print_run_list(run_list)

    # do this after config.handle_standard_args, as command line args may override injectables
    injectables = ["data_dir", "configs_dir", "output_dir", "settings_file_name"]
    injectables = {k: inject.get_injectable(k) for k in injectables}

    mp_tasks.run_multiprocess(run_list, injectables)
    pipeline.open_pipeline("_")
    regress_3_zone()
    pipeline.close_pipeline()


if __name__ == "__main__":

    test_mp_run()



# ActivitySim
# See full license in LICENSE.txt

import logging

import numpy as np
import pandas as pd

from activitysim.abm.models.util import estimation
from activitysim.core import chunk, config, inject, logit, pipeline, tracing
from activitysim.core.util import reindex

from .util import probabilistic_scheduling as ps

logger = logging.getLogger(__name__)


def run_tour_scheduling_probabilistic(
    tours_df,
    scheduling_probs,
    probs_join_cols,
    depart_alt_base,
    chunk_size,
    trace_label,
    trace_hh_id,
):
    """Make probabilistic tour scheduling choices in chunks

    Parameters
    ----------
    tours_df : pandas.DataFrame
        table of tours
    scheduling_probs : pandas.DataFrame
        Probability lookup table for tour depature and return times
    probs_join_cols : str or list of strs
        Columns to use for merging probability lookup table with tours table
    depart_alt_base : int
        int to add to probs column index to get time period it represents.
        e.g. depart_alt_base = 5 means first column (column 0) represents 5 am
    chunk_size : int
        size of chooser chunks, set in main settings.yaml
    trace_label : str
        label to append to tracing logs and table names
    trace_hh_id : int
        households to trace

    Returns
    -------
    pandas.Series
        series of chosen alternative indices for each chooser
    """
    result_list = []
    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        tours_df, chunk_size, trace_label, trace_label
    ):
        choices = ps.make_scheduling_choices(
            chooser_chunk,
            "departure",
            scheduling_probs,
            probs_join_cols,
            depart_alt_base,
            first_trip_in_leg=False,
            report_failed_trips=True,
            trace_label=chunk_trace_label,
            trace_hh_id=trace_hh_id,
            trace_choice_col_name="depart_return",
            clip_earliest_latest=False,
        )
        result_list.append(choices)

    choices = pd.concat(result_list)
    return choices


@inject.step()
def tour_scheduling_probabilistic(tours, chunk_size, trace_hh_id):
    """Makes tour departure and arrival choices by sampling from a probability lookup table

    This model samples tour scheduling choices from an exogenously defined probability
    distribution rather than simulating choices from a discrete choice model. This is particularly
    useful when estimating from sparse survey data with small numbers of observations
    across tour scheduling alternatives.

    Parameters
    ----------
    tours :  orca.DataFrameWrapper
        lazy-loaded table of tours
    chunk_size :  int
        size of chooser chunks, defined in main settings.yaml
    trace_hh_id : int
        households to trace, defined in main settings.yaml

    """

    trace_label = "tour_scheduling_probabilistic"
    model_settings_file_name = "tour_scheduling_probabilistic.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)
    depart_alt_base = model_settings.get("depart_alt_base", 0)
    scheduling_probs_filepath = config.config_file_path(model_settings["PROBS_SPEC"])
    scheduling_probs = pd.read_csv(scheduling_probs_filepath)
    probs_join_cols = model_settings["PROBS_JOIN_COLS"]
    tours_df = tours.to_frame()

    # trip_scheduling is a probabilistic model ane we don't support estimation,
    # but we do need to override choices in estimation mode
    estimator = estimation.manager.begin_estimation("tour_scheduling_probabilistic")
    if estimator:
        estimator.write_spec(model_settings, tag="PROBS_SPEC")
        estimator.write_model_settings(model_settings, model_settings_file_name)
        chooser_cols_for_estimation = ["purpose_id"]
        estimator.write_choosers(tours_df[chooser_cols_for_estimation])

    choices = run_tour_scheduling_probabilistic(
        tours_df,
        scheduling_probs,
        probs_join_cols,
        depart_alt_base,
        chunk_size,
        trace_label,
        trace_hh_id,
    )

    # convert alt index choices to depart/return times
    probs_cols = pd.Series(
        [c for c in scheduling_probs.columns if c not in probs_join_cols]
    )
    dep_ret_choices = probs_cols.loc[choices]
    dep_ret_choices.index = choices.index
    choices.update(dep_ret_choices)
    departures = choices.str.split("_").str[0].astype(int)
    returns = choices.str.split("_").str[1].astype(int)

    if estimator:
        estimator.write_choices(choices)
        choices_df = estimator.get_survey_values(
            choices, "tours", ["start", "end"]
        )  # override choices
        choices = choices_df["start"].astype(str) + "_" + choices_df["end"].astype(str)
        estimator.write_override_choices(choices)
        estimator.end_estimation()
        assert not choices.isnull().any()

    # these column names are required for downstream models (e.g. tour mode choice)
    # normally generated by time_windows.py and used as alts for vectorize_tour_scheduling
    tours_df["start"] = departures
    tours_df["end"] = returns
    tours_df["duration"] = tours_df["end"] - tours_df["start"]

    assert not tours_df["start"].isnull().any()
    assert not tours_df["end"].isnull().any()
    assert not tours_df["duration"].isnull().any()

    pipeline.replace_table("tours", tours_df)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, inject, pipeline, simulate, tracing
from activitysim.core.util import assign_in_place

from .util import estimation, tour_od

logger = logging.getLogger(__name__)


@inject.step()
def tour_od_choice(
    tours, persons, households, land_use, network_los, chunk_size, trace_hh_id
):

    """Simulates joint origin/destination choice for all tours.

    Given a set of previously generated tours, each tour needs to have an
    origin and a destination. In this case tours are the choosers, but
    the associated person that's making the tour does not necessarily have
    a home location assigned already. So we choose a tour origin at the same
    time as we choose a tour destination, and assign the tour origin as that
    person's home location.

    Parameters
    ----------
    tours : orca.DataFrameWrapper
        lazy-loaded tours table
    persons : orca.DataFrameWrapper
        lazy-loaded persons table
    households : orca.DataFrameWrapper
        lazy-loaded households table
    land_use : orca.DataFrameWrapper
        lazy-loaded land use data table
    stop_frequency_alts : orca.DataFrameWrapper
        lazy-loaded table of stop frequency alternatives, e.g. "1out2in"
    network_los : orca._InjectableFuncWrapper
        lazy-loaded activitysim.los.Network_LOS object
    chunk_size
        simulation chunk size, set in main settings.yaml
    trace_hh_id : int
        households to trace, set in main settings.yaml
    """

    trace_label = "tour_od_choice"
    model_settings_file_name = "tour_od_choice.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)
    origin_col_name = model_settings["ORIG_COL_NAME"]
    dest_col_name = model_settings["DEST_COL_NAME"]
    alt_id_col = tour_od.get_od_id_col(origin_col_name, dest_col_name)

    sample_table_name = model_settings.get("OD_CHOICE_SAMPLE_TABLE_NAME")
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    logsum_column_name = model_settings.get("OD_CHOICE_LOGSUM_COLUMN_NAME", None)
    want_logsums = logsum_column_name is not None

    tours = tours.to_frame()

    # interaction_sample_simulate insists choosers appear in same order as alts
    tours = tours.sort_index()

    estimator = estimation.manager.begin_estimation("tour_od_choice")
    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        estimator.write_spec(model_settings, tag="SAMPLE_SPEC")
        estimator.write_spec(model_settings, tag="SPEC")
        estimator.set_alt_id(alt_id_col)
        estimator.write_table(
            inject.get_injectable("size_terms"), "size_terms", append=False
        )
        estimator.write_table(
            inject.get_table("land_use").to_frame(), "landuse", append=False
        )
        estimator.write_model_settings(model_settings, model_settings_file_name)

    choices_df, save_sample_df = tour_od.run_tour_od(
        tours,
        persons,
        want_logsums,
        want_sample_table,
        model_settings,
        network_los,
        estimator,
        chunk_size,
        trace_hh_id,
        trace_label,
    )

    if estimator:
        assert estimator.want_unsampled_alternatives
        estimator.write_choices(choices_df.choice)
        survey_od = estimator.get_survey_values(
            choices_df.choice, "tours", [origin_col_name, dest_col_name]
        )
        choices_df[origin_col_name] = survey_od[origin_col_name]
        choices_df[dest_col_name] = survey_od[dest_col_name]
        survey_od[alt_id_col] = tour_od.create_od_id_col(
            survey_od, origin_col_name, dest_col_name
        )
        choices_df.choice = survey_od[alt_id_col]
        estimator.write_override_choices(choices_df.choice)
        estimator.end_estimation()

    tours[origin_col_name] = choices_df[origin_col_name].reindex(tours.index)
    tours[dest_col_name] = choices_df[dest_col_name].reindex(tours.index)
    if want_logsums:
        tours[logsum_column_name] = (
            choices_df["logsum"].reindex(tours.index).astype("float")
        )
    tours["poe_id"] = tours[origin_col_name].map(
        land_use.to_frame(columns="poe_id").poe_id
    )

    households = households.to_frame()
    persons = persons.to_frame()
    households[origin_col_name] = tours.set_index("household_id")[
        origin_col_name
    ].reindex(households.index)
    persons[origin_col_name] = (
        households[origin_col_name].reindex(persons.household_id).values
    )

    # Downstream steps require that persons and households have a 'home_zone_id'
    # column. We assume that if the tour_od_choice model is used, this field is
    # missing from the population data, so it gets inherited from the tour origin
    households["home_zone_id"] = households[origin_col_name]
    persons["home_zone_id"] = persons[origin_col_name]

    pipeline.replace_table("tours", tours)
    pipeline.replace_table("persons", persons)
    pipeline.replace_table("households", households)

    if want_sample_table:
        assert len(save_sample_df.index.get_level_values(0).unique()) == len(choices_df)
        pipeline.extend_table(sample_table_name, save_sample_df)

    if trace_hh_id:
        tracing.trace_df(
            tours,
            label="tours_od_choice",
            slicer="person_id",
            index_label="tour",
            columns=None,
            warn_if_empty=True,
        )



import argparse
import os

import numpy as np
import openmatrix as omx
import pandas as pd

MAZ_OFFSET = 0

segments = {
    "test": {"MAZ": np.arange(MAZ_OFFSET + 492, MAZ_OFFSET + 1101)},  # includes univ
    "univ_east": {"MAZ": np.arange(MAZ_OFFSET, MAZ_OFFSET + 1080)},
    "full": {},
}

parser = argparse.ArgumentParser(description="crop SANDAG 3 zone raw_data")
parser.add_argument(
    "segment_name",
    metavar="segment_name",
    type=str,
    nargs=1,
    help=f"geography segmentation (e.g. full)",
)

parser.add_argument(
    "-c",
    "--check_geography",
    default=False,
    action="store_true",
    help="check consistency of MAZ, TAZ, TAP zone_ids and foreign keys & write orphan_households file",
)

args = parser.parse_args()


segment_name = args.segment_name[0]
check_geography = args.check_geography

assert segment_name in segments.keys(), f"Unknown seg: {segment_name}"

input_dir = "./data_raw"
output_dir = f"./data_{segment_name}_3"


print(f"segment_name {segment_name}")

print(f"input_dir {input_dir}")
print(f"output_dir {output_dir}")

print(f"check_geography {check_geography}")

if not os.path.isdir(output_dir):
    print(f"creating output directory {output_dir}")
    os.mkdir(output_dir)


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def patch_maz(df, maz_offset):
    for c in df.columns:
        if c in ["MAZ", "OMAZ", "DMAZ", "mgra", "orig_mgra", "dest_mgra"]:
            df[c] += maz_offset
    return df


def read_csv(file_name):
    df = pd.read_csv(input_path(file_name))
    if MAZ_OFFSET:
        df = patch_maz(df, MAZ_OFFSET)
    print(f"read {file_name} {df.shape}")
    return df


def to_csv(df, file_name):
    df.to_csv(output_path(file_name), index=False)
    print(f"write {file_name} {df.shape}")


def crop_omx(omx_file_name, zones, num_outfiles=1):

    skim_data_type = np.float32

    omx_in = omx.open_file(input_path(f"{omx_file_name}.omx"))
    print(f"omx_in shape {omx_in.shape()}")

    offset_map = None
    for mapping_name in omx_in.listMappings():
        _offset_map = np.asanyarray(omx_in.mapentries(mapping_name))
        offset_map = _offset_map

    om = pd.Series(offset_map)
    om = om[om.isin(zones.values)]
    indexes = om.index.values

    labels = zones.values  # TAZ zone_ids in omx index order

    # create
    if num_outfiles == 1:
        omx_out = [omx.open_file(output_path(f"{omx_file_name}.omx"), "w")]
    else:
        omx_out = [
            omx.open_file(output_path(f"{omx_file_name}{i + 1}.omx"), "w")
            for i in range(num_outfiles)
        ]

    for omx_file in omx_out:
        omx_file.create_mapping("ZONE", labels)

    iskim = 0
    for mat_name in omx_in.list_matrices():
        # make sure we have a vanilla numpy array, not a CArray
        m = np.asanyarray(omx_in[mat_name]).astype(skim_data_type)
        m = m[indexes, :][:, indexes]
        print(f"{mat_name} {m.shape}")

        omx_file = omx_out[iskim % num_outfiles]
        omx_file[mat_name] = m
        iskim += 1

    omx_in.close()
    for omx_file in omx_out:
        omx_file.close()


# non-standard input file names

LAND_USE = "land_use.csv"
HOUSEHOLDS = "households.csv"
PERSONS = "persons.csv"
MAZ_TAZ = "maz.csv"
TAP_MAZ = "tap.csv"
TAZ = "taz.csv"


if check_geography:

    # ######## check for orphan_households not in any maz in land_use
    land_use = read_csv(LAND_USE)
    land_use = land_use[["MAZ", "TAZ"]]
    land_use = land_use.sort_values(["TAZ", "MAZ"])

    households = read_csv(HOUSEHOLDS)
    orphan_households = households[~households.MAZ.isin(land_use.MAZ)]
    print(f"{len(orphan_households)} orphan_households")

    # write orphan_households to INPUT directory (since it doesn't belong in output)
    if len(orphan_households) > 0:
        file_name = "orphan_households.csv"
        print(
            f"writing {file_name} {orphan_households.shape} to {input_path(file_name)}"
        )
        orphan_households.to_csv(input_path(file_name), index=False)

    # ######## check that land_use and maz and taz tables have same MAZs and TAZs

    # could just build maz and taz files, but want to make sure PSRC data is right

    land_use = read_csv(LAND_USE)
    # assert land_use.set_index('MAZ').index.is_monotonic_increasing

    land_use = land_use.sort_values("MAZ")
    maz = read_csv(MAZ_TAZ).sort_values("MAZ")

    # ### FATAL ###
    if not land_use.MAZ.isin(maz.MAZ).all():
        print(
            f"land_use.MAZ not in maz.MAZ\n{land_use.MAZ[~land_use.MAZ.isin(maz.MAZ)]}"
        )
        raise RuntimeError(f"land_use.MAZ not in maz.MAZ")

    if not maz.MAZ.isin(land_use.MAZ).all():
        print(f"maz.MAZ not in land_use.MAZ\n{maz.MAZ[~maz.MAZ.isin(land_use.MAZ)]}")

    # ### FATAL ###
    if not land_use.TAZ.isin(maz.TAZ).all():
        print(
            f"land_use.TAZ not in maz.TAZ\n{land_use.TAZ[~land_use.TAZ.isin(maz.TAZ)]}"
        )
        raise RuntimeError(f"land_use.TAZ not in maz.TAZ")

    if not maz.TAZ.isin(land_use.TAZ).all():
        print(f"maz.TAZ not in land_use.TAZ\n{maz.TAZ[~maz.TAZ.isin(land_use.TAZ)]}")


# land_use

land_use = read_csv(LAND_USE)

land_use.MAZ = land_use.MAZ.astype(int)

ur_land_use = land_use.copy()

slicer = segments[segment_name]
for slice_col, slice_values in slicer.items():
    # print(f"slice {slice_col}: {slice_values}")
    land_use = land_use[land_use[slice_col].isin(slice_values)]

print(f"land_use shape after slicing {land_use.shape}")
to_csv(land_use, "land_use.csv")


# TAZ

taz = pd.DataFrame({"TAZ": sorted(ur_land_use.TAZ.unique())})
taz = taz[taz.TAZ.isin(land_use["TAZ"])]
to_csv(taz, TAZ)


# maz_taz


maz_taz = read_csv(MAZ_TAZ).sort_values("MAZ")
maz_taz = maz_taz[maz_taz.MAZ.isin(land_use.MAZ)]
to_csv(maz_taz, MAZ_TAZ)

# tap

taps = read_csv(TAP_MAZ)
taps = taps[["TAP", "MAZ"]].sort_values(by="TAP").reset_index(drop=True)
taps = taps[taps["MAZ"].isin(land_use["MAZ"])]
to_csv(taps, "tap.csv")

# maz to tap

maz_tap_walk = read_csv("maz_to_tap_walk.csv").sort_values(["MAZ", "TAP"])
taz_tap_drive = read_csv("maz_to_tap_drive.csv").sort_values(["MAZ", "TAP"])

maz_tap_walk = maz_tap_walk[
    maz_tap_walk["MAZ"].isin(land_use["MAZ"]) & maz_tap_walk["TAP"].isin(taps["TAP"])
]
taz_tap_drive = taz_tap_drive[
    taz_tap_drive["MAZ"].isin(land_use["MAZ"]) & taz_tap_drive["TAP"].isin(taps["TAP"])
]

to_csv(maz_tap_walk, "maz_to_tap_walk.csv")
to_csv(taz_tap_drive, "maz_to_tap_drive.csv")

# maz to mz

maz_maz_walk = read_csv("maz_to_maz_walk.csv").sort_values(["OMAZ", "DMAZ"])
maz_maz_bike = read_csv("maz_to_maz_bike.csv").sort_values(["OMAZ", "DMAZ"])

maz_maz_walk = maz_maz_walk[
    maz_maz_walk["OMAZ"].isin(land_use["MAZ"])
    & maz_maz_walk["DMAZ"].isin(land_use["MAZ"])
]
maz_maz_bike = maz_maz_bike[
    maz_maz_bike["OMAZ"].isin(land_use["MAZ"])
    & maz_maz_bike["DMAZ"].isin(land_use["MAZ"])
]

to_csv(maz_maz_walk, "maz_to_maz_walk.csv")
to_csv(maz_maz_bike, "maz_to_maz_bike.csv")

# tap_lines

tap_lines = read_csv("tap_lines.csv")
tap_lines = tap_lines[tap_lines["TAP"].isin(taps["TAP"])]
to_csv(tap_lines, "tap_lines.csv")

# households

households = read_csv(HOUSEHOLDS)
households = households[households["MAZ"].isin(land_use["MAZ"])]
to_csv(households, "households.csv")

# persons

persons = read_csv(PERSONS)
persons = persons[persons["household_id"].isin(households["HHID"])]
to_csv(persons, "persons.csv")

# skims

crop_omx("taz_skims1", taz.TAZ, num_outfiles=(4 if segment_name == "full" else 1))
crop_omx("tap_skims1", taps.TAP, num_outfiles=(4 if segment_name == "full" else 1))
